/*  $Header: /var/lib/cvs/prism54-ng/ksrc/isl_ioctl.c,v 1.123 2004/01/28 22:07:09 jmaurer Exp $
 *  
 *  Copyright (C) 2002 Intersil Americas Inc.
 *            (C) 2003 Aurelien Alleaume <slts@free.fr>
 *            (C) 2003 Herbert Valerio Riedel <hvr@gnu.org>
 *            (C) 2003 Luis R. Rodriguez <mcgrof@ruslug.rutgers.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/if_arp.h>
#include <linux/pci.h>

#include <asm/uaccess.h>

#include "isl_ioctl.h"
#include "islpci_mgt.h"
#include "isl_oid.h"		/* additional types and defs for isl38xx fw */

#undef IW_MAX_BITRATES
#define	IW_MAX_BITRATES	16

#if WIRELESS_EXT > 12
#include <net/iw_handler.h>	/* New driver API */
#endif				/* WIRELESS_EXT > 12 */

/****************************************************************************
 * common patterns 
 */

/* set patterns */

static int
_mgt_set_request(islpci_private *priv, u32 oid, void *data, unsigned dlen,
		 void *cache)
{
	int ret = 0;
	struct islpci_mgmtframe *response;
	int response_op = PIMFOR_OP_ERROR;

	if (cache)
		if (down_interruptible(&priv->mib.sem))
			return -ERESTARTSYS;

	if (islpci_get_state(priv) >= PRV_STATE_INIT) {
		ret = islpci_mgt_transaction(priv->ndev, PIMFOR_OP_SET, oid,
					     data, dlen, &response);
		if (!ret) {
			response_op = response->header->operation;
			islpci_mgt_release(response);
		}
		if (ret || response_op == PIMFOR_OP_ERROR)
		        ret = -EINVAL;
	}

	if (cache) {
		if (!ret)
			memcpy(cache, data, dlen);
		up(&priv->mib.sem);
	}

	return ret;
}

static inline int
_mgt_set_request_u32(islpci_private *priv, u32 oid, u32 datum, void *cache)
{
	u32 _datum = cpu_to_le32(datum);
	return _mgt_set_request(priv, oid, &_datum, sizeof (_datum), cache);
}

/* The NOCACHE version of the following macros are allowed in both READY and 
 * INIT state. And they don't save anything to mib (thus they don't take its
 * lock).
 */

#define MGT_TEMPLATE(x) do { \
	   int _rvalue = x; \
           if (_rvalue) return _rvalue;    \
	   } while (0)

#define MGT_SET_REQUEST(priv,oid,data,dlen,name) \
           MGT_TEMPLATE(_mgt_set_request ((priv), (oid), (data), (dlen), \
                                  &(priv)->mib.name))

#define MGT_SET_REQUEST_NOCACHE(priv,oid,data,dlen)  \
            MGT_TEMPLATE(_mgt_set_request ((priv), (oid), (data), (dlen), NULL))

#define MGT_SET_REQUEST_NOCACHE_NORET(priv,oid,data,dlen)  \
            _mgt_set_request ((priv), (oid), (data), (dlen), NULL)

#define MGT_SET_REQUEST_U32(priv,oid,datum,name) \
            MGT_TEMPLATE(_mgt_set_request_u32 ((priv), (oid), (datum), \
                                 (void*) &(priv)->mib.name)) \

#define MGT_SET_REQUEST_U32_NOCACHE(priv,oid,datum) \
            MGT_TEMPLATE(_mgt_set_request_u32 ((priv), (oid), (datum), NULL))

/* get patterns */

static int
_mgt_get_request(islpci_private *priv, u32 oid, void *data, void **res,
		 unsigned dlen, void *cache, struct islpci_mgmtframe **response)
{
	int ret = -EINVAL;
	int reslen = 0;

	*response = NULL;
	*res = NULL;
	if (cache)
		if (down_interruptible(&priv->mib.sem))
			return -ERESTARTSYS;

	if (islpci_get_state(priv) >= PRV_STATE_INIT) {
		ret = islpci_mgt_transaction(priv->ndev, PIMFOR_OP_GET,
					     oid, data, dlen, response);
		if (ret || !*response ||
			(*response)->header->operation == PIMFOR_OP_ERROR) {
			if (*response)
				islpci_mgt_release(*response);
			ret = -EINVAL;
		}
		if (!ret) {
			*res = (*response)->data;
			reslen = (*response)->header->length;
		}
	} else if (cache) {
		*res = cache;
		reslen = dlen;
		ret = 0;
	}

	if (cache)
		up(&(priv)->mib.sem);

	/* The following situation is normal whith some oids for which we don't
	 * know exactly how much data the card will send back. eg : when we get
	 * the BSS list.
	 */
	if (!ret && (reslen != dlen))
		printk(KERN_DEBUG
		       "_mgt_get_request(): data length (%d!=%d) mismatch for oid=0x%x\n",
		       reslen, dlen, oid);

	return ret;
}

static inline int
_mgt_get_request_u32(islpci_private *priv, u32 oid, u32 *datum, void *cache)
{
	u32 *result = 0;
	struct islpci_mgmtframe *frame;

	int ret = _mgt_get_request(priv, oid, NULL, (void **) &result,
				       sizeof (u32), cache, &frame);
	if (ret)
		return ret;
	*datum = le32_to_cpu(*result);
	if (frame)
		islpci_mgt_release(frame);

	return 0;
}

#define MGT_GET_REQUEST(priv,oid,data,res,dlen,name,entry)  \
            MGT_TEMPLATE(_mgt_get_request ((priv), (oid), (data),	\
		                (res), (dlen), &((priv)->mib.name), (entry)))

#define MGT_GET_REQUEST_NOCACHE(priv,oid,data,res,dlen,entry)  \
            MGT_TEMPLATE(_mgt_get_request ((priv), (oid), (data),	\
		                   (res), (dlen), NULL, (entry)))	\

#define MGT_FREE_REQUEST(priv,entry) do { \
	if ((entry))			\
		islpci_mgt_release((entry)); 	\
	} while (0)

#define MGT_GET_REQUEST_U32(priv,oid,datum,name)   \
            MGT_TEMPLATE(_mgt_get_request_u32 ((priv), (oid), (datum), \
                                 (void*) &((priv)->mib.name))) \

#define MGT_GET_REQUEST_U32_NOCACHE(priv,oid,datum)  \
            MGT_TEMPLATE(_mgt_get_request_u32 ((priv), (oid), (datum), NULL))

/****************************************************************************/

static int init_mode = CARD_DEFAULT_IW_MODE;
static int init_channel = CARD_DEFAULT_CHANNEL;
static int init_wep = CARD_DEFAULT_WEP;
static int init_filter = CARD_DEFAULT_FILTER;
static int init_authen = CARD_DEFAULT_AUTHEN;
static int init_dot1x = CARD_DEFAULT_DOT1X;
static int init_conformance = CARD_DEFAULT_CONFORMANCE;
static int init_mlme = CARD_DEFAULT_MLME_MODE;

MODULE_PARM(init_mode, "i");
MODULE_PARM_DESC(init_mode,
		 "Set card mode:\n0: Auto\n1: Ad-Hoc\n2: Managed Client (Default)\n3: Master / Access Point\n4: Repeater (Not supported yet)\n5: Secondary (Not supported yet)\n6: Monitor");

MODULE_PARM(init_channel, "i");
MODULE_PARM_DESC(init_channel,
		 "Check `iwpriv ethx channel` for available channels");

MODULE_PARM(init_wep, "i");
MODULE_PARM(init_filter, "i");

MODULE_PARM(init_authen, "i");
MODULE_PARM_DESC(init_authen,
		 "Authentication method. Can be of seven types:\n0 0x0000: None\n1 0x0001: DOT11_AUTH_OS (Default)\n2 0x0002: DOT11_AUTH_SK\n3 0x0003: DOT11_AUTH_BOTH");

MODULE_PARM(init_dot1x, "i");
MODULE_PARM_DESC(init_dot1x,
		 "\n0: None/not set	(Default)\n1: DOT11_DOT1X_AUTHENABLED\n2: DOT11_DOT1X_KEYTXENABLED");

MODULE_PARM(init_mlme, "i");
MODULE_PARM_DESC(init_mlme,
		 "Sets the MAC layer management entity (MLME) mode of operation,\n0: DOT11_MLME_AUTO (Default)\n1: DOT11_MLME_INTERMEDIATE\n2: DOT11_MLME_EXTENDED");

#ifdef CONFIG_PRISM54_WDS
int init_wds = CARD_DEFAULT_WDS;
MODULE_PARM(init_wds, "i");
#endif

/**
 * prism54_mib_mode_helper - MIB change mode helper function
 * @mib: the &struct islpci_mib object to modify
 * @iw_mode: new mode (%IW_MODE_*)
 * 
 *  This is a helper function, hence it does not lock. Make sure
 *  caller deals with locking *if* necessary. This function sets the 
 *  mode-dependent mib values and does the mapping of the Linux 
 *  Wireless API modes to Device firmware modes. It also checks for 
 *  correct valid Linux wireless modes. 
 */
int
prism54_mib_mode_helper(struct islpci_mib *mib, u32 *iw_mode)
{

	/* Let's see if the user passed a valid Linux Wireless mode */
	if (*iw_mode > IW_MODE_MONITOR || *iw_mode < IW_MODE_AUTO) {
		printk(KERN_DEBUG
		       "%s(): You passed a non-valid Linux Wireless Mode.\n",
		       __FUNCTION__);
		return -1;
	}

	/* For now, just catch early the Repeater and Secondary modes here */
	if (*iw_mode == IW_MODE_REPEAT || *iw_mode == IW_MODE_SECOND) {
		printk(KERN_DEBUG "%s(): Sorry, Repeater mode and Secondary mode \
				are not yet supported by this driver.\n",
		       __FUNCTION__);
		return -1;
	}

	/* Default mib values or cleaning out of mode-depending mibs which may affect
	 * new mode operations */
	mib->mlmeautolevel = CARD_DEFAULT_MLME_MODE;
	mib->bsstype = CARD_DEFAULT_BSSTYPE;
	mib->iw_mode = (u32) *iw_mode;
	mib->config = INL_CONFIG_MANUALRUN & ~INL_CONFIG_RXANNEX;

	/* Note: To report full 802.11 frames in monitor mode we need to enable
	 * INL_CONFIG_TXANNEX in OID_INL_CONFIG. */
	switch (*iw_mode) {
	case IW_MODE_AUTO:
		mib->mode = INL_MODE_CLIENT;
		mib->bsstype = DOT11_BSSTYPE_ANY;
		break;
	case IW_MODE_ADHOC:
		mib->mode = INL_MODE_CLIENT;
		mib->bsstype = DOT11_BSSTYPE_IBSS;
		break;
	case IW_MODE_INFRA:
		mib->mode = INL_MODE_CLIENT;
		mib->bsstype = DOT11_BSSTYPE_INFRA;
		break;
	case IW_MODE_MASTER:
		mib->mode = INL_MODE_AP;
		mib->bsstype = DOT11_BSSTYPE_INFRA;
		break;
	case IW_MODE_MONITOR:
		mib->mode = INL_MODE_PROMISCUOUS;
		mib->bsstype = DOT11_BSSTYPE_ANY;
		mib->config =
		    INL_CONFIG_MANUALRUN | INL_CONFIG_RXANNEX |
		    INL_CONFIG_TXANNEX;
		break;
	case IW_MODE_REPEAT:	/* or */
	case IW_MODE_SECOND:	/* Not supported yet. */
		printk(KERN_DEBUG "Sorry, Repeater mode and Secondary \
					mode not yet supported\n");
	default:		/* Yes, the breaks above were left out on purpose */
		/* Mode passed is incorrect, let's set it to the old
		 * the mode */
		return -1;
		break;
	}
	return 0;		/* Succcess */
}

/**
 * prism54_mib_init_defaults - fill MIB cache with defaults
 * @mib: the &struct islpci_mib to initialize
 *
 *  this function initializes the struct given as @mib with defaults,
 *  of which many are retrieved from the global module parameter
 *  variables.  
 */

void
prism54_mib_init_defaults(struct islpci_mib *mib)
{
	int idx;
	char *model;
	char **modelp;

	memset(mib, 0, sizeof *mib);
	sema_init(&mib->sem, 1);

	/* no need to lock here */
	
	model = "PRISM Duette/GT"; /* Default */
	modelp = &model;

	/* ----- Set standard mib values ----- */
	mib->iw_mode = init_mode;
	mib->channel = init_channel;
	mib->authenable = init_authen;
	mib->privacyinvoked = init_wep;
	mib->exunencrypted = init_filter;
	mib->dot1xenable = init_dot1x;
	mib->mlmeautolevel = init_mlme;
	mib->conformance = init_conformance;

	for (idx = 0; idx < 4; ++idx) {
		mib->defkey[idx].type = DOT11_PRIV_WEP;
		mib->defkey[idx].length = 0;
	}
	/* ----- End of standard mib values ----- */
	/* ----- Now, let's deal with the mode-dependent mib values ----- */
	mib->config = INL_CONFIG_NOTHING;
	if (init_wds)
		mib->config |= INL_CONFIG_WDS;
	/* Important: we are setting a default wireless mode and we are 
	 * forcing a valid one, so prism54_mib_mode_helper should just set
	 * mib values depending on what the wireless mode given is. No need
	 * for it save old values */
	if (init_mode > IW_MODE_MONITOR || init_mode < IW_MODE_AUTO) {
		printk(KERN_DEBUG "%s(): You passed a non-valid init_mode. \
				Using default mode\n", __FUNCTION__);
		init_mode = CARD_DEFAULT_IW_MODE;
	}
	/* This sets all of the mode-dependent values */
	prism54_mib_mode_helper(mib, (u32 *) &init_mode);
	/* ----- End of mode dependent values ----- */
#if 0
	/* One last mode dependent value, but that only needs to be set at init time */
	if (init_mode == IW_MODE_MASTER) {
		mib->ssid.length = sizeof (CARD_DEFAULT_AP_SSID) - 1;
		memcpy(mib->ssid.octets, CARD_DEFAULT_AP_SSID,
		       mib->ssid.length);
	} else {
		mib->ssid.length = sizeof (CARD_DEFAULT_CLIENT_SSID) - 1;
		memcpy(mib->ssid.octets, CARD_DEFAULT_CLIENT_SSID,
		       mib->ssid.length);
	}
	/* Now let's actually SET the values */
#endif
}


/* If you want to lock the mib during the commit, acquire and release it out
 * this function because the MGT_SET_REQUEST_* macros can return...
 */
int
prism54_mib_init(islpci_private *priv)
{
	struct obj_buffer psm_buffer = {
		.size = cpu_to_le32(PSM_BUFFER_SIZE),
		.addr = cpu_to_le32(priv->device_psm_buffer)
	};

	/* update ndev's MAC address from firmware */
	{
		void *mac_addr = 0;
		struct islpci_mgmtframe *entry = 0;

		MGT_GET_REQUEST_NOCACHE(priv, GEN_OID_MACADDRESS, NULL,
					&mac_addr, 6, &entry);
		memcpy(priv->ndev->dev_addr, mac_addr, 6);
		MGT_FREE_REQUEST(priv, entry);
	}

	/* reset config mode */
	MGT_SET_REQUEST_U32_NOCACHE(priv, OID_INL_CONFIG, priv->mib.config);

	/* Set the mode a first time regardless of the config. If config had the 
	 * CONFIG_MANUAL_RUN bit on, the device will be stopped at the first mode set.
	 * We then have to set all desirable custom values since the device will use the
	 * default ones for the device otherwise. A second mode set will be required at 
	 * the end, to make the device go on using the new pre-set changes for its mode 
	 * (turn the device on using those changes). 
	 *
	 * Without the CONFIG_MANUAL_RUN, all mode sets will start out fresh
	 * with new default settings. Note that for Linux we need CONFIG_MANUAL_RUN for 
	 * all modes except IW_MODE_AUTO */

	/* set the mode a first time in order to stop the device */
	MGT_SET_REQUEST_U32_NOCACHE(priv, OID_INL_MODE, priv->mib.mode);

	/* set the bsstype */
	MGT_SET_REQUEST_U32_NOCACHE(priv, DOT11_OID_BSSTYPE, priv->mib.bsstype);

	/* set the channel/frequency */
	if (priv->mib.frequency)
		MGT_SET_REQUEST_U32_NOCACHE(priv, DOT11_OID_FREQUENCY,
					    priv->mib.frequency);
	else			/* frequency == 0 -> use channel value */
		MGT_SET_REQUEST_U32_NOCACHE(priv, DOT11_OID_CHANNEL,
					    priv->mib.channel);
	{
		char *data;
                struct islpci_mgmtframe *entry;
		MGT_GET_REQUEST_NOCACHE(priv, OID_INL_VERSION, NULL,
					(void **) &data, 8, &entry);
		printk(KERN_DEBUG "%s: firmware version %s\n",
		       priv->ndev->name, data);
		MGT_FREE_REQUEST(priv, entry);
	}

	/* set the mlme mode object */
	MGT_SET_REQUEST_U32_NOCACHE(priv, DOT11_OID_MLMEAUTOLEVEL,
				    priv->mib.mlmeautolevel);

	if (priv->mib.mode == INL_MODE_PROMISCUOUS) {
		/* set the mode again for actually starting the device */
		MGT_SET_REQUEST_U32_NOCACHE(priv, OID_INL_MODE, priv->mib.mode);
		printk(KERN_ALERT
		       "%s: done with prism54_mib_init()! - Monitor mode!\n",
		       priv->ndev->name);
		return 0;
	}

	/* set the SSID */
	MGT_SET_REQUEST_NOCACHE(priv, DOT11_OID_SSID, (void *) &priv->mib.ssid,
				sizeof (struct obj_ssid));

	/* set the PSM buffer object */
	MGT_SET_REQUEST_NOCACHE(priv, DOT11_OID_PSMBUFFER,
				(void *) &psm_buffer,
				sizeof (struct obj_buffer));

	/* set the authentication enable option */
	MGT_SET_REQUEST_U32_NOCACHE(priv, DOT11_OID_AUTHENABLE,
				    priv->mib.authenable);

	/* set the privacy invoked mode */
	MGT_SET_REQUEST_U32_NOCACHE(priv, DOT11_OID_PRIVACYINVOKED,
				    priv->mib.privacyinvoked);

	/* set the exunencrypted object */
	MGT_SET_REQUEST_U32_NOCACHE(priv, DOT11_OID_EXUNENCRYPTED,
				    priv->mib.exunencrypted);

	/* set default key 1-4 */
	MGT_SET_REQUEST_NOCACHE(priv, DOT11_OID_DEFKEY1,
				&priv->mib.defkey[0], sizeof (struct obj_key));

	MGT_SET_REQUEST_NOCACHE(priv, DOT11_OID_DEFKEY2,
				&priv->mib.defkey[1], sizeof (struct obj_key));

	MGT_SET_REQUEST_NOCACHE(priv, DOT11_OID_DEFKEY3,
				&priv->mib.defkey[2], sizeof (struct obj_key));

	MGT_SET_REQUEST_NOCACHE(priv, DOT11_OID_DEFKEY4,
				&priv->mib.defkey[3], sizeof (struct obj_key));

	/* set the dot1x enable object */
	MGT_SET_REQUEST_U32_NOCACHE(priv, DOT11_OID_DOT1XENABLE,
				    priv->mib.dot1xenable);

	/* set the 802.11d conformance mode object  */
	MGT_SET_REQUEST_U32_NOCACHE(priv, OID_INL_DOT11D_CONFORMANCE,
				    priv->mib.conformance);

	/* set the mode again for actually starting the device */
	MGT_SET_REQUEST_U32_NOCACHE(priv, OID_INL_MODE, priv->mib.mode);

	printk(KERN_DEBUG "%s: done with %s()! -- Client, Managed, Master mode\n",
	       priv->ndev->name, __FUNCTION__);

	return 0;
}

void
prism54_mib_init_work(islpci_private *priv)
{
	if (down_interruptible(&priv->mib.sem))
		return;
	prism54_mib_init(priv);
	up(&priv->mib.sem);
}

int
prism54_stop(islpci_private *priv)
{
	/* Don't cache the following set requests because we might want to
	 * restart the device without loosing all the config.
	 */

	MGT_SET_REQUEST_U32_NOCACHE(priv, OID_INL_CONFIG, 0);

	MGT_SET_REQUEST_U32_NOCACHE(priv, OID_INL_MODE, INL_MODE_NONE);

	return 0;
}

static int
prism54_update_stats_helper(islpci_private *priv)
{
	struct islpci_mgmtframe *entry;
	char *data;
	u32 u;
	int j;
	struct obj_bss bss, *bss2;

/* missing stats are :
 *  iwstatistics.qual.updated
 *  iwstatistics.discard.nwid	    
 *  iwstatistics.discard.fragment	    
 *  iwstatistics.discard.misc
 *  iwstatistics.miss.beacon */

/* Noise floor.
 * I'm not sure if the unit is dBm.
 * Note : If we are not connected, this value seems to be irrevelant. */

	MGT_GET_REQUEST_U32_NOCACHE(priv, DOT11_OID_NOISEFLOOR, &u);
	priv->local_iwstatistics.qual.noise = u;

/* Get the rssi of the link. To do this we need to retrieve a bss. */

	/* First get the MAC address of the AP we are associated with. */
	MGT_GET_REQUEST_NOCACHE(priv, DOT11_OID_BSSID, NULL,
				(void **) &data, 6, &entry);

	/* copy this MAC to the bss */
	for (j = 0; j < 6; j++)
		bss.address[j] = data[j];
	MGT_FREE_REQUEST(priv, entry);

	/* now ask for the corresponding bss */
	MGT_GET_REQUEST_NOCACHE(priv, DOT11_OID_BSSFIND, (void *) &bss,
				(void **) &bss2, sizeof (struct obj_bss),
				&entry);

	/* report the rssi and use it to calculate
	 *  link quality through a signal-noise
	 *  ratio */
	priv->local_iwstatistics.qual.level = bss2->rssi;
	priv->local_iwstatistics.qual.qual =
	    bss2->rssi - priv->iwstatistics.qual.noise;

	MGT_FREE_REQUEST(priv, entry);

	/* report that the stats are new */
	priv->local_iwstatistics.qual.updated = 0x7;

/* Rx : unable to decrypt the MPDU */
	MGT_GET_REQUEST_U32_NOCACHE(priv, DOT11_OID_PRIVRXFAILED, &u);
	priv->local_iwstatistics.discard.code = u;

/* Tx : Max MAC retries num reached */
	MGT_GET_REQUEST_U32_NOCACHE(priv, DOT11_OID_MPDUTXFAILED, &u);
	priv->local_iwstatistics.discard.retries = u;

	return 0;
}

/* this will be executed outside of atomic context thanks to
 * schedule_work(), thus we can as well use sleeping semaphore
 * locking */
int
prism54_update_stats(islpci_private *priv)
{
	int rvalue;

	if (down_interruptible(&priv->stats_sem))
		return -ERESTARTSYS;
	rvalue = prism54_update_stats_helper(priv);
	up(&priv->stats_sem);

	return rvalue;
}

struct iw_statistics *
prism54_get_wireless_stats(struct net_device *ndev)
{
	islpci_private *priv = ndev->priv;

	/* If the stats are being updated return old data */
	if (down_trylock(&priv->stats_sem) == 0) {
		memcpy(&priv->iwstatistics, &priv->local_iwstatistics,
		       sizeof (struct iw_statistics));
		/* They won't be marked updated for the next time */
		priv->local_iwstatistics.qual.updated = 0;
		up(&priv->stats_sem);
	} else
		priv->iwstatistics.qual.updated = 0;

	/* Update our wireless stats, but do not schedule to often 
	 * (max 1 HZ) */
	if ((priv->stats_timestamp == 0) ||
	    time_after(jiffies, priv->stats_timestamp + 1 * HZ)) {
		schedule_work(&priv->stats_work);
		priv->stats_timestamp = jiffies;
	}

	return &priv->iwstatistics;
}

static int
prism54_commit(struct net_device *ndev, struct iw_request_info *info,
	       char *cwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	/* printk(KERN_DEBUG "%s: SIOCSIWCOMMIT called\n", ndev->name); */

	/* if the card is not ready, we should return succes. */
	if (islpci_get_state(priv) != PRV_STATE_READY)
		return 0;

	/* simply re-set the last set SSID, this should commit most stuff */
	MGT_SET_REQUEST_NOCACHE(priv, DOT11_OID_SSID, &priv->mib.ssid,
				sizeof (struct obj_ssid));

	return 0;
}

static int
prism54_get_name(struct net_device *ndev, struct iw_request_info *info,
		 char *cwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	char *capabilities;
	u32 cap;

	if (islpci_get_state(priv) < PRV_STATE_INIT){
		strncpy(cwrq, "NOT READY!", IFNAMSIZ);
		return 0;
	}

	MGT_GET_REQUEST_U32_NOCACHE(ndev->priv, OID_INL_PHYCAPABILITIES, &cap);
	switch (cap)
	{
		case INL_PHYCAP_5000MHZ:
			capabilities = "IEEE 802.11a/b/g";
			break;
		case INL_PHYCAP_FAA:
			capabilities = "IEEE 802.11b/g - FAA Support";
			break;
		case INL_PHYCAP_2400MHZ:
		default:
			capabilities = "IEEE 802.11b/g";	/* Default */
			break;
	}
	strncpy(cwrq, capabilities, IFNAMSIZ);
	return 0;
}

static int
prism54_set_freq(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_freq *fwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	int mantisse, exponent;

	/* prepare the structure for the set object */
	if (fwrq->m < 1000) {
		/* structure value contains a channel indication */
		MGT_SET_REQUEST_U32(priv, DOT11_OID_CHANNEL, fwrq->m, channel);
		priv->mib.frequency = 0;
	} else {
		/* structure contains a frequency indication */
		/* mantisse is larger than 1000, so make it kHz */
		mantisse = fwrq->m / 1000;
		exponent = fwrq->e;

		/* process the exponent */
		while (exponent > 0) {
			mantisse *= 10;
			exponent--;
		}

		BUG_ON(exponent != 0);
		MGT_SET_REQUEST_U32(priv, DOT11_OID_FREQUENCY, mantisse,
				    frequency);
		priv->mib.channel = 0;
	}

	/* Call commit handler */
	return -EINPROGRESS;
}

static int
prism54_get_freq(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_freq *fwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	if (priv->mib.frequency) {
		/* the user wants to use frequency value */
		MGT_GET_REQUEST_U32(priv, DOT11_OID_FREQUENCY, &fwrq->m,
				    frequency);
		fwrq->e = 3;
	} else {
		/* use channel number */
		MGT_GET_REQUEST_U32(priv, DOT11_OID_CHANNEL, &fwrq->m, channel);
		fwrq->e = 0;
	}

	return 0;
}

static int
prism54_set_mode(struct net_device *ndev, struct iw_request_info *info,
		 __u32 * uwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	/* Knock knock */
	if (down_interruptible(&priv->mib.sem))
		/* No one's home :( */
		return -ERESTARTSYS;

	/* Let's see if the user passed a valid Linux Wireless mode */
	if (*uwrq > IW_MODE_MONITOR || *uwrq < IW_MODE_AUTO) {
		printk(KERN_DEBUG
		       "%s: %s() You passed a non-valid init_mode.\n",
		       priv->ndev->name, __FUNCTION__);
		up(&priv->mib.sem);
		return -EOPNOTSUPP;
	}

	if (prism54_mib_mode_helper(&priv->mib, uwrq)) {
		up(&priv->mib.sem);
		return -EOPNOTSUPP;
	}

	/* the ACL code needs an intermediate mlmeautolevel */
	if ((*uwrq == IW_MODE_MASTER) && (priv->acl.policy != MAC_POLICY_OPEN))
		priv->mib.mlmeautolevel = DOT11_MLME_INTERMEDIATE;
	if (priv->wpa)
		priv->mib.mlmeautolevel = DOT11_MLME_EXTENDED;

	/* update MIB copy now that requests succeeded
	 * Note : there is a small window  for race here : 
	 * if the acl code replies to some message while priv->mib.mode=master
	 * and the card is in clent mode, it will hang the card ! */
	prism54_mib_init(priv);
	priv->ndev->type = (priv->mib.iw_mode == IW_MODE_MONITOR)
	    ? ARPHRD_IEEE80211 : ARPHRD_ETHER;
	up(&priv->mib.sem);
	return 0;
}

/* Use mib cache */
static int
prism54_get_mode(struct net_device *ndev, struct iw_request_info *info,
		 __u32 * uwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	struct islpci_mib *mib = &priv->mib;
	/* We could use BUG_ON here but not now.. BUG_ON can lockup hard */
	if ((mib->iw_mode < IW_MODE_AUTO) || (mib->iw_mode > IW_MODE_MONITOR))
		printk(KERN_DEBUG "BUG found on %s\n", __FUNCTION__);
	*uwrq = mib->iw_mode;
	return 0;
}

/* we use DOT11_OID_EDTHRESHOLD. From what I guess the card will not try to
 * emit data if (sensitivity > rssi - noise) (in dBm).
 * prism54_set_sens does not seem to work.
 */

static int
prism54_set_sens(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 sens;

	/* by default  the card sets this to 20. */
	sens = vwrq->disabled ? 20 : vwrq->value;

	/* set the ed threshold. */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_EDTHRESHOLD, sens, edthreshold);

	return 0;
}

static int
prism54_get_sens(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 sens;

	/* request the ed threshold. */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_EDTHRESHOLD, &sens, edthreshold);

	/* return the data */
	vwrq->value = sens;
	vwrq->disabled = (vwrq->value == 0);
	vwrq->fixed = 1;

	return 0;
}

static int
prism54_get_range(struct net_device *ndev, struct iw_request_info *info,
		  struct iw_point *dwrq, char *extra)
{
	struct iw_range *range = (struct iw_range *) extra;
	islpci_private *priv = ndev->priv;
	struct islpci_mgmtframe *entry;
	char *data;
	int i, m;
	struct obj_frequencies *freq;

	/* set the wireless extension version number */
	range->we_version_source = SUPPORTED_WIRELESS_EXT;
	range->we_version_compiled = WIRELESS_EXT;

	/* Now the encoding capabilities */
	range->num_encoding_sizes = 3;
	/* 64(40) bits WEP */
	range->encoding_size[0] = 5;
	/* 128(104) bits WEP */
	range->encoding_size[1] = 13;
	/* 256 bits for WPA-PSK */
	range->encoding_size[2] = 32;
	/* 4 keys are allowed */
	range->max_encoding_tokens = 4;

	/* we don't know the quality range... */
	range->max_qual.level = 0;
	range->max_qual.noise = 0;
	range->max_qual.qual = 0;
	/* these value describe an average quality. Needs more tweaking... */
	range->avg_qual.level = -80;	/* -80 dBm */
	range->avg_qual.noise = 0;	/* don't know what to put here */
	range->avg_qual.qual = 0;

	range->sensitivity = 200;

	/* retry limit capabilities */
	range->retry_capa = IW_RETRY_LIMIT | IW_RETRY_LIFETIME;
	range->retry_flags = IW_RETRY_LIMIT;
	range->r_time_flags = IW_RETRY_LIFETIME;

	/* I don't know the range. Put stupid things here */
	range->min_retry = 1;
	range->max_retry = 65535;
	range->min_r_time = 1024;
	range->max_r_time = 65535 * 1024;

	/* txpower is supported in dBm's */
	range->txpower_capa = IW_TXPOW_DBM;

	if (islpci_get_state(priv) < PRV_STATE_INIT)
		return 0;

	/* Request the device for the supported frequencies
	 * not really revelant since some devices will report the 5 GHz band
	 * frequencies even if they don't support them.
	 */
#define IWMAXFREQ	32
	MGT_GET_REQUEST_NOCACHE(priv, DOT11_OID_SUPPORTEDFREQUENCIES, NULL,
				(void **) &freq,
				(IWMAXFREQ + 1) * sizeof (short) + 2, &entry);

	range->num_channels = le16_to_cpu(freq->nr);
	range->num_frequency = le16_to_cpu(freq->nr);

	/* Frequencies are not listed in the right order. The reordering is probably
	 * firmware dependant and thus should work for everyone.
	 */
	m = min(IW_MAX_FREQUENCIES, (int) freq->nr);
	for (i = 0; i < m - 12; i++) {
		range->freq[i].m = le16_to_cpu(freq->mhz[12 + i]);
		range->freq[i].e = 6;
		range->freq[i].i = i + 1;
	}
	for (i = m - 12; i < m; i++) {
		range->freq[i].m = le16_to_cpu(freq->mhz[i - m + 12]);
		range->freq[i].e = 6;
		range->freq[i].i = i + 23;
	}

	MGT_FREE_REQUEST(priv, entry);

	/* request the device for the suppoted rates */
	MGT_GET_REQUEST_NOCACHE(priv, DOT11_OID_SUPPORTEDRATES, NULL,
				(void **) &data,
				(IW_MAX_BITRATES + 1) * sizeof (char), &entry);

	/* We got an array of char. It is NULL terminated. */
	i = 0;
	while ((i < IW_MAX_BITRATES) && (*data != 0)) {
		/* the result must be in bps. The card gives us 500Kbps */
		range->bitrate[i] = (__s32) (*data >> 1);
		range->bitrate[i] *= 1000000;
		i++;
		data++;
	}

	range->num_bitrates = i;
	MGT_FREE_REQUEST(priv, entry);

	return 0;
}

/* Set AP address*/

static int
prism54_set_wap(struct net_device *ndev, struct iw_request_info *info,
		struct sockaddr *awrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	char bssid[6];

	if (awrq->sa_family != ARPHRD_ETHER)
		return -EINVAL;

	/* prepare the structure for the set object */
	memcpy(&bssid[0], awrq->sa_data, 6);

	/* set the bssid -- does this make sense when in AP mode? */
	MGT_SET_REQUEST(priv, DOT11_OID_BSSID, &bssid, 6, bssid);

	return -EINPROGRESS;	/* Call commit handler */
}

/* get AP address*/

static int
prism54_get_wap(struct net_device *ndev, struct iw_request_info *info,
		struct sockaddr *awrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	struct islpci_mgmtframe *entry;
	void *data;

	/* request the device for the essid */
	MGT_GET_REQUEST(priv, DOT11_OID_BSSID, NULL, (void **) &data, 6,
			bssid, &entry);

	/* return the data */
	memcpy(awrq->sa_data, data, 6);
	awrq->sa_family = ARPHRD_ETHER;

	MGT_FREE_REQUEST(priv, entry);

	return 0;
}

static int
prism54_set_scan(struct net_device *dev, struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{
	/* hehe the device does this automagicaly */
	return 0;
}

/* a little helper that will translate our data into a card independent
 * format that the Wireless Tools will understand. This was inspired by
 * the "Aironet driver for 4500 and 4800 series cards" (GPL)
 */

inline char *
prism54_translate_bss(struct net_device *ndev, char *current_ev,
		      char *end_buf, struct obj_bss *bss, char noise)
{
	struct iw_event iwe;	/* Temporary buffer */
	short cap;
	islpci_private *priv = ndev->priv;

	/* The first entry must be the MAC address */
	memcpy(iwe.u.ap_addr.sa_data, bss->address, 6);
	iwe.u.ap_addr.sa_family = ARPHRD_ETHER;
	iwe.cmd = SIOCGIWAP;
	current_ev =
	    iwe_stream_add_event(current_ev, end_buf, &iwe, IW_EV_ADDR_LEN);

	/* The following entries will be displayed in the same order we give them */

	/* The ESSID. It's prefixed by its length (one char) */
	iwe.u.data.length = bss->ssid.length;
	iwe.u.data.flags = 1;
	iwe.cmd = SIOCGIWESSID;
	current_ev = iwe_stream_add_point(current_ev, end_buf,
					  &iwe, bss->ssid.octets);

	/* Capabilities */
#define CAP_ESS 0x01
#define CAP_IBSS 0x02
#define CAP_CRYPT 0x10

	/* Mode */
	cap = le16_to_cpu(bss->capinfo);
	iwe.u.mode = 0;
	if (cap & CAP_ESS)
		iwe.u.mode = IW_MODE_MASTER;
	else if (cap & CAP_IBSS)
		iwe.u.mode = IW_MODE_ADHOC;
	iwe.cmd = SIOCGIWMODE;
	if (iwe.u.mode)
		current_ev =
		    iwe_stream_add_event(current_ev, end_buf, &iwe,
					 IW_EV_UINT_LEN);

	/* Encryption capability */
	if (cap & CAP_CRYPT)
		iwe.u.data.flags = IW_ENCODE_ENABLED | IW_ENCODE_NOKEY;
	else
		iwe.u.data.flags = IW_ENCODE_DISABLED;
	iwe.u.data.length = 0;
	iwe.cmd = SIOCGIWENCODE;
	current_ev = iwe_stream_add_point(current_ev, end_buf, &iwe, NULL);

	/* Add frequency. (short) bss->channel is the frequency in MHz */
	iwe.u.freq.m = bss->channel;
	iwe.u.freq.e = 6;
	iwe.cmd = SIOCGIWFREQ;
	current_ev =
	    iwe_stream_add_event(current_ev, end_buf, &iwe, IW_EV_FREQ_LEN);

	/* Add quality statistics */
	iwe.u.qual.level = bss->rssi;
	iwe.u.qual.noise = noise;
	/* do a simple SNR for quality */
	iwe.u.qual.qual = bss->rssi - noise;
	iwe.cmd = IWEVQUAL;
	current_ev =
	    iwe_stream_add_event(current_ev, end_buf, &iwe, IW_EV_QUAL_LEN);

#if WIRELESS_EXT > 14
	if (priv->wpa) {
		u8 wpa_ie[MAX_WPA_IE_LEN];
		char *buf, *p;
		size_t wpa_ie_len;
		int i;

		wpa_ie_len = prism54_wpa_ie_get(priv, bss->address, wpa_ie);
		if (wpa_ie_len > 0 &&
		    (buf = kmalloc(wpa_ie_len * 2 + 10, GFP_ATOMIC))) {
			p = buf;
			p += sprintf(p, "wpa_ie=");
			for (i = 0; i < wpa_ie_len; i++) {
				p += sprintf(p, "%02x", wpa_ie[i]);
			}
			memset(&iwe, 0, sizeof(iwe));
			iwe.cmd = IWEVCUSTOM;
			iwe.u.data.length = strlen(buf);
			current_ev = iwe_stream_add_point(current_ev, end_buf,
							  &iwe, buf);
			kfree(buf);
		}
	}
#endif /* WIRELESS_EXT > 14 */

	return current_ev;
}

int
prism54_get_scan(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	struct islpci_mgmtframe *entry;
	int i;
	struct obj_bsslist *bsslist;
	u32 noise = 0;
	char *current_ev = extra;

	if (islpci_get_state(priv) < PRV_STATE_INIT) {
		/* device is not ready, fail gently */
		dwrq->length = 0;
		return 0;
	}

	/* first get the noise value. We will use it to report the link quality */
	MGT_GET_REQUEST_U32_NOCACHE(priv, DOT11_OID_NOISEFLOOR, &noise);

	/* Ask the device for a list of known bss. We can report at most
	 * IW_MAX_AP=64 to the range struct. But the device won't repport anything
	 * if you change the value of MAXBSS=24. Anyway 24 AP It is probably enough.
	 */
#define IWMAXBSS	24

	MGT_GET_REQUEST_NOCACHE(priv, DOT11_OID_BSSLIST, NULL,
				(void **) &bsslist,
				8 + 4 + IWMAXBSS * sizeof (struct obj_bss),
				&entry);

	/* ok now, scan the list and translate its info */
	for (i = 0; i < min(IW_MAX_AP, (int) le32_to_cpu(bsslist->nr)); i++)
		current_ev = prism54_translate_bss(ndev, current_ev,
						   extra + IW_SCAN_MAX_DATA,
						   &(bsslist->bsslist[i]),
						   noise);
	MGT_FREE_REQUEST(priv, entry);
	dwrq->length = (current_ev - extra);
	dwrq->flags = 0;	/* todo */

	return 0;
}

static int
prism54_set_essid(struct net_device *ndev, struct iw_request_info *info,
		  struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	struct obj_ssid essid_obj;

	memset(essid_obj.octets, 0, 33);

	/* Check if we were asked for `any' */
	if (dwrq->flags) {
		/* we must set the SSID. */
		if (dwrq->length > min(33, IW_ESSID_MAX_SIZE + 1))
			return -E2BIG;
		essid_obj.length = dwrq->length - 1;
		memcpy(essid_obj.octets, extra, essid_obj.length);
	} else {
		/* we must clear the ssid */
		essid_obj.length = 0;
	}

	/* request the device to set the essid */
	MGT_SET_REQUEST(priv, DOT11_OID_SSID, &essid_obj,
			sizeof (struct obj_ssid), ssid);

	return 0;
}

static int
prism54_get_essid(struct net_device *ndev, struct iw_request_info *info,
		  struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	struct islpci_mgmtframe *entry;
	struct obj_ssid *essid_obj;

	/* request the device for the essid */
	MGT_GET_REQUEST(priv, DOT11_OID_SSID, NULL,
			(void **) &essid_obj, sizeof (struct obj_ssid), ssid,
			&entry);

	/* copy the data to the return structure */
	dwrq->flags = 1;	/* set ESSID to ON for Wireless Extensions */
	dwrq->length = essid_obj->length + 1;

	memcpy(extra, essid_obj->octets, dwrq->length);

	MGT_FREE_REQUEST(priv, entry);

	return 0;
}

/* Provides no functionality, just completes the ioctl. In essence this is a 
 * just a cosmetic ioctl.
 */
static int
prism54_set_nick(struct net_device *ndev, struct iw_request_info *info,
		struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	if (dwrq->length > IW_ESSID_MAX_SIZE)
		return -E2BIG;

	if (down_interruptible(&priv->mib.sem))
		return -ERESTARTSYS;

	memset(priv->mib.nick, 0, sizeof(priv->mib.nick));
	memcpy(priv->mib.nick, extra, dwrq->length);

	up(&priv->mib.sem);
	return 0;
}

static int
prism54_get_nick(struct net_device *ndev, struct iw_request_info *info,
		struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	dwrq->length = 0;
	
	if (down_interruptible(&priv->mib.sem))
		return -ERESTARTSYS;
	dwrq->length = strlen(priv->mib.nick) + 1;
	memcpy(extra, priv->mib.nick, dwrq->length);
	up(&priv->mib.sem);
	return 0;
}

/* Set the allowed Bitrates - BROKEN - disabled */

static int
prism54_set_rate(struct net_device *ndev,
		 struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{

	islpci_private *priv = ndev->priv;
	u32 rate;

	if ((vwrq->value < 8) && (vwrq->value >= 0))
		/* it is a rate index */
		rate = vwrq->value;
	else
		rate = (u32) (vwrq->value / 500000);

	if (vwrq->value == -1) {
		/* auto mode. No limit. */

	}

	/* Now, check if we want a fixed or auto value */
	if (vwrq->fixed == 0) {
		/* set the maximum allowed bitrate */

	} else {
		/* Fixed rate  mode */

	}

	/* For now, always set the fixed rate mode. Other mode are harder
	 * to implement.
	 */

	MGT_SET_REQUEST_U32(priv, DOT11_OID_ALOFT_FIXEDRATE, rate, fixedrate);

	return 0;
}

/* Get the current bit rate */

static int
prism54_get_rate(struct net_device *ndev,
		 struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 u;

	/* Get the current bit rate */
	MGT_GET_REQUEST_U32_NOCACHE(priv, GEN_OID_LINKSTATE, &vwrq->value);
	vwrq->value *= 500000;

	/* request the device for the fixed rate */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_ALOFT_FIXEDRATE, &u, fixedrate);

	/* if *data = -1 we are not in fixed rate mode. Report it. */
	vwrq->fixed = (u != -1);

	return 0;
}

static int
prism54_set_rts(struct net_device *ndev, struct iw_request_info *info,
		struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	/* set the rts threshold */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_RTSTHRESH, vwrq->value, rtsthresh);

	return 0;
}

static int
prism54_get_rts(struct net_device *ndev, struct iw_request_info *info,
		struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 u;

	/* get the rts threshold */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_RTSTHRESH, &u, rtsthresh);
	vwrq->value = u;

	return 0;
}

static int
prism54_set_frag(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	/*  set the fragment threshold */
	MGT_SET_REQUEST_U32(priv, DOT11_OID_FRAGTHRESH, vwrq->value,
			    fragthresh);

	return 0;
}

static int
prism54_get_frag(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 u;

	/* get the fragment threshold */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_FRAGTHRESH, &u, fragthresh);
	vwrq->value = u;

	return 0;
}

/* Here we have (min,max) = max retries for (small frames, big frames). Where
 * big frame <=>  bigger than the rts threshold
 * small frame <=>  smaller than the rts threshold
 * This is not really the behavior expected by the wireless tool but it seems
 * to be a common behavior in other drivers.
 * 
 * It seems that playing with this tends to hang the card -> DISABLED
 */

static int
prism54_set_retry(struct net_device *ndev, struct iw_request_info *info,
		  struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	int slimit = 0, llimit = 0;	/* short and long limit */
	int lifetime = 0;

	if (vwrq->disabled)
		/* we cannot disable this feature */
		return -EINVAL;

	if (vwrq->flags & IW_RETRY_LIMIT) {
		if (vwrq->flags & IW_RETRY_MIN)
			slimit = vwrq->value;
		else if (vwrq->flags & IW_RETRY_MAX)
			llimit = vwrq->value;
		else {
			/* we are asked to set both */
			slimit = vwrq->value;
			llimit = vwrq->value;
		}
	}
	if (vwrq->flags & IW_RETRY_LIFETIME)
		/* Wireless tools use us unit while the device uses 1024 us unit */
		lifetime = vwrq->value / 1024;

	/* now set what is requested */

	if (slimit != 0)
		MGT_SET_REQUEST_U32(priv, DOT11_OID_SHORTRETRIES, slimit,
				    shortretries);
	if (llimit != 0)
		MGT_SET_REQUEST_U32(priv, DOT11_OID_LONGRETRIES, llimit,
				    longretries);
	if (lifetime != 0)
		MGT_SET_REQUEST_U32(priv, DOT11_OID_MAXTXLIFETIME,
				    lifetime, maxtxlifetime);

	return 0;
}

static int
prism54_get_retry(struct net_device *ndev, struct iw_request_info *info,
		  struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 u;

	vwrq->disabled = 0;	/* It cannot be disabled */

	if ((vwrq->flags & IW_RETRY_TYPE) == IW_RETRY_LIFETIME) {
		/* we are asked for the life time */
		MGT_GET_REQUEST_U32(priv, DOT11_OID_MAXTXLIFETIME, &u,
				    maxtxlifetime);
		vwrq->value = u * 1024;
		vwrq->flags = IW_RETRY_LIFETIME;
	} else if ((vwrq->flags & IW_RETRY_MAX)) {
		/* we are asked for the long retry limit */
		MGT_GET_REQUEST_U32(priv, DOT11_OID_LONGRETRIES, &u,
				    longretries);
		vwrq->value = u;
		vwrq->flags = IW_RETRY_LIMIT | IW_RETRY_MAX;
	} else {
		/* default. get the  short retry limit */
		MGT_GET_REQUEST_U32(priv, DOT11_OID_SHORTRETRIES, &u,
				    shortretries);
		vwrq->value = u;
		vwrq->flags = IW_RETRY_LIMIT | IW_RETRY_MIN;
	}

	return 0;
}

/* Now the WEP functions. We begin with 2 helpers. */

static int
getdefkeyx(islpci_private *priv, struct obj_key *res, u32 index)
{
	struct obj_key *pkey;
	u32 oid;
	struct islpci_mgmtframe *entry;

	switch (index) {
	case 0:
		oid = DOT11_OID_DEFKEY1;
		break;

	case 1:
		oid = DOT11_OID_DEFKEY2;
		break;

	case 2:
		oid = DOT11_OID_DEFKEY3;
		break;

	case 3:
		oid = DOT11_OID_DEFKEY4;
		break;

	default:
		return -EOPNOTSUPP;
	}

	/* get the corresponding key */
	MGT_GET_REQUEST(priv, oid, NULL, (void **) &pkey,
			sizeof (struct obj_key), defkey[index], &entry);
	memcpy(res, pkey, sizeof (struct obj_key));
	MGT_FREE_REQUEST(priv, entry);
	return 0;
}

static int
setdefkeyx(islpci_private *priv, struct obj_key *key, u32 index)
{
	u32 oid;

	switch (index) {
	case 0:
		oid = DOT11_OID_DEFKEY1;
		break;

	case 1:
		oid = DOT11_OID_DEFKEY2;
		break;

	case 2:
		oid = DOT11_OID_DEFKEY3;
		break;

	case 3:
		oid = DOT11_OID_DEFKEY4;
		break;

	default:
		return -EOPNOTSUPP;
	}

	/* set the corresponding key */
	MGT_SET_REQUEST(priv, oid, key, sizeof (struct obj_key), defkey[index]);

	return 0;
}

/* And now the real functions */

static int
prism54_set_encode(struct net_device *ndev, struct iw_request_info *info,
		   struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	int rvalue;
	int force = 0;

	int authen = DOT11_AUTH_OS;
	int invoke = 0;
	int exunencrypt = 0;

	/* with the new API, it's impossible to get a NULL pointer.
	 * New version of iwconfig set the IW_ENCODE_NOKEY flag
	 * when no key is given, but older versions don't. */

	if (dwrq->length > 0) {
		/* we have a key to set */
		int index = (dwrq->flags & IW_ENCODE_INDEX) - 1;
		int current_index;
		struct obj_key key = { DOT11_PRIV_WEP, 0, "" };

		/* get the current key index */
		MGT_GET_REQUEST_U32(priv, DOT11_OID_DEFKEYID,
				    &current_index, defkeyid);
		/* Verify that the key is not marked as invalid */
		if (!(dwrq->flags & IW_ENCODE_NOKEY)) {
			key.length = dwrq->length > sizeof (key.key) ?
			    sizeof (key.key) : dwrq->length;
			memcpy(key.key, extra, key.length);
			if (key.length == 32)
				/* we want WPA-PSK */
				key.type = DOT11_PRIV_TKIP;
			if ((index < 0) || (index > 3))
				/* no index provided use the current one */
				index = current_index;

			/* now send the key to the card  */
			if (rvalue = setdefkeyx(priv, &key, index), rvalue != 0)
				return rvalue;
		}
		/*
		 * If a valid key is set, encryption should be enabled 
		 * (user may turn it off later).
		 * This is also how "iwconfig ethX key on" works
		 */
		if ((index == current_index) && (key.length > 0))
			force = 1;
	} else {
		int index = (dwrq->flags & IW_ENCODE_INDEX) - 1;
		if ((index >= 0) && (index <= 3)) {
			/* we want to set the key index */
			MGT_SET_REQUEST_U32(priv, DOT11_OID_DEFKEYID,
					    index, defkeyid);
		} else {
			if (!dwrq->flags & IW_ENCODE_MODE) {
				/* we cannot do anything. Complain. */
				return -EINVAL;
			}
		}
	}

	/* now read the flags     */
	if (dwrq->flags & IW_ENCODE_DISABLED) {
		/* Encoding disabled, 
		 * authen = DOT11_AUTH_OS;
		 * invoke = 0;
		 * exunencrypt = 0; */
	}
	if ((dwrq->flags & IW_ENCODE_OPEN) || force)
		/* Encode but accept non-encoded packets. No auth */
		invoke = 1;
	if ((dwrq->flags & IW_ENCODE_RESTRICTED) || force) {
		/* Refuse non-encoded packets. Auth */
		authen = DOT11_AUTH_BOTH;
		invoke = 1;
		exunencrypt = 1;
	}
	/* do the change if requested  */
	if ((dwrq->flags & IW_ENCODE_MODE) || force) {
		MGT_SET_REQUEST_U32(priv, DOT11_OID_AUTHENABLE, authen,
				    authenable);
		MGT_SET_REQUEST_U32(priv, DOT11_OID_PRIVACYINVOKED, invoke,
				    privacyinvoked);
		MGT_SET_REQUEST_U32(priv, DOT11_OID_EXUNENCRYPTED, exunencrypt,
				    exunencrypted);
	}
	return 0;
}

static int
prism54_get_encode(struct net_device *ndev, struct iw_request_info *info,
		   struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	struct obj_key key;
	u32 devindex, index = (dwrq->flags & IW_ENCODE_INDEX) - 1;
	u32 authen = 0, invoke = 0, exunencrypt = 0;

	/* first get the flags */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_AUTHENABLE, &authen, authenable);
	MGT_GET_REQUEST_U32(priv, DOT11_OID_PRIVACYINVOKED, &invoke,
			    privacyinvoked);
	MGT_GET_REQUEST_U32(priv, DOT11_OID_EXUNENCRYPTED, &exunencrypt,
			    exunencrypted);

	if (invoke && (authen == DOT11_AUTH_BOTH) && exunencrypt)
		dwrq->flags = IW_ENCODE_RESTRICTED;
	else if ((authen == DOT11_AUTH_OS) && !exunencrypt) {
		if (invoke)
			dwrq->flags = IW_ENCODE_OPEN;
		else
			dwrq->flags = IW_ENCODE_DISABLED;
	} else
		/* The card should not work in this state */
		dwrq->flags = 0;

	/* get the current device key index */
	MGT_GET_REQUEST_U32(priv, DOT11_OID_DEFKEYID, &devindex, defkeyid);

	/* Now get the key, return it */
	if ((index < 0) || (index > 3))
		/* no index provided, use the current one */
		index = devindex;
	if (!getdefkeyx(priv, &key, index)) {
		dwrq->length = key.length;
		memcpy(extra, key.key, dwrq->length);
	} else
		dwrq->length = 0;
	/* return the used key index */
	dwrq->flags |= devindex + 1;
	return 0;
}



static int
prism54_get_txpower(struct net_device *ndev, struct iw_request_info *info,
		struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	u32 u;
	memset(vwrq, 0, sizeof(struct iw_param));
	MGT_GET_REQUEST_U32(priv, OID_INL_OUTPUTPOWER, &u, fixedtxpower);
	// intersil firmware operates in 0.25 dBm (1/4 dBm)
	vwrq->value = u / 4;
	vwrq->fixed = 1;	// fixed value is set
	// radio is not turned of
	// btw: how is possible to turn off only the radio ?
	vwrq->disabled = 0; 
	return 0;
}	//	isl_ioctl_gettxpow

static int 
prism54_set_txpower(struct net_device *ndev, struct iw_request_info *info,
		struct iw_param *vwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	// intersil firmware operates in 0.25 dBm (1/4)
	u32 u = vwrq->value;
	u *= 4;
	if (vwrq->disabled) {
		// don't know how to disable radio
		printk(KERN_DEBUG "%s: %s() disabling radio is not yet supported.\n", priv->ndev->name, __FUNCTION__);
		return -ENOTSUPP;
	} else if (vwrq->fixed)	{
		// currently only fixed value is supported
		MGT_SET_REQUEST_U32(priv, OID_INL_OUTPUTPOWER , u, fixedtxpower);
		return 0;
	} else {
		printk(KERN_DEBUG "%s: %s() auto power will be implemented later.\n", 
				priv->ndev->name, __FUNCTION__);
		return -ENOTSUPP;
	}
}	//	isl_ioctl_settxpow



static int
prism54_reset(struct net_device *ndev, struct iw_request_info *info,
	      __u32 * uwrq, char *extra)
{
	islpci_reset(ndev->priv, 0);

	return 0;
}

static int
prism54_set_beacon(struct net_device *ndev, struct iw_request_info *info,
		   __u32 * uwrq, char *extra)
{
	MGT_SET_REQUEST_U32((islpci_private *) ndev->priv,
			    DOT11_OID_BEACONPERIOD, *uwrq, beaconperiod);

	return -EINPROGRESS;
}

static int
prism54_get_beacon(struct net_device *ndev, struct iw_request_info *info,
		   __u32 * uwrq, char *extra)
{
	MGT_GET_REQUEST_U32((islpci_private *) ndev->priv,
			    DOT11_OID_BEACONPERIOD, uwrq, beaconperiod);

	return 0;
}

void
prism54_acl_init(struct islpci_acl *acl)
{
	sema_init(&acl->sem, 1);
	INIT_LIST_HEAD(&acl->mac_list);
	acl->size = 0;
	acl->policy = MAC_POLICY_OPEN;
}

static void
prism54_clear_mac(struct islpci_acl *acl)
{
	struct list_head *ptr, *next;
	struct mac_entry *entry;

	if (down_interruptible(&acl->sem))
		return;

	if (acl->size == 0) {
		up(&acl->sem);
		return;
	}

	for (ptr = acl->mac_list.next, next = ptr->next;
	     ptr != &acl->mac_list; ptr = next, next = ptr->next) {
		entry = list_entry(ptr, struct mac_entry, _list);
		list_del(ptr);
		kfree(entry);
	}
	acl->size = 0;
	up(&acl->sem);
}

void
prism54_acl_clean(struct islpci_acl *acl)
{
	prism54_clear_mac(acl);
}

static int
prism54_add_mac(struct net_device *ndev, struct iw_request_info *info,
		struct sockaddr *awrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	struct islpci_acl *acl = &priv->acl;
	struct mac_entry *entry;
	struct sockaddr *addr = (struct sockaddr *) extra;

	if (addr->sa_family != ARPHRD_ETHER)
		return -EOPNOTSUPP;

	entry = kmalloc(sizeof (struct mac_entry), GFP_KERNEL);
	if (entry == NULL)
		return -ENOMEM;

	memcpy(entry->addr, addr->sa_data, ETH_ALEN);

	if (down_interruptible(&acl->sem)) {
		kfree(entry);
		return -ERESTARTSYS;
	}
	list_add_tail(&entry->_list, &acl->mac_list);
	acl->size++;
	up(&acl->sem);

	return 0;
}

static int
prism54_del_mac(struct net_device *ndev, struct iw_request_info *info,
		struct sockaddr *awrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	struct islpci_acl *acl = &priv->acl;
	struct mac_entry *entry;
	struct list_head *ptr;
	struct sockaddr *addr = (struct sockaddr *) extra;

	if (addr->sa_family != ARPHRD_ETHER)
		return -EOPNOTSUPP;

	if (down_interruptible(&acl->sem))
		return -ERESTARTSYS;
	for (ptr = acl->mac_list.next; ptr != &acl->mac_list; ptr = ptr->next) {
		entry = list_entry(ptr, struct mac_entry, _list);

		if (memcmp(entry->addr, addr->sa_data, ETH_ALEN) == 0) {
			list_del(ptr);
			acl->size--;
			kfree(entry);
			up(&acl->sem);
			return 0;
		}
	}
	up(&acl->sem);
	return -EINVAL;
}

static int
prism54_get_mac(struct net_device *ndev, struct iw_request_info *info,
		struct iw_point *dwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	struct islpci_acl *acl = &priv->acl;
	struct mac_entry *entry;
	struct list_head *ptr;
	struct sockaddr *dst = (struct sockaddr *) extra;

	dwrq->length = 0;

	if (down_interruptible(&acl->sem))
		return -ERESTARTSYS;

	for (ptr = acl->mac_list.next; ptr != &acl->mac_list; ptr = ptr->next) {
		entry = list_entry(ptr, struct mac_entry, _list);

		memcpy(dst->sa_data, entry->addr, ETH_ALEN);
		dst->sa_family = ARPHRD_ETHER;
		dwrq->length++;
		dst++;
	}
	up(&acl->sem);
	return 0;
}

/* Setting policy also clears the MAC acl, even if we don't change the defaut
 * policy
 */

static int
prism54_set_policy(struct net_device *ndev, struct iw_request_info *info,
		   __u32 * uwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	struct islpci_acl *acl = &priv->acl;

	prism54_clear_mac(acl);

	if ((*uwrq < MAC_POLICY_OPEN) || (*uwrq > MAC_POLICY_REJECT))
		return -EINVAL;

	if (down_interruptible(&priv->mib.sem))
		return -ERESTARTSYS;

	acl->policy = *uwrq;

	/* the ACL code needs an intermediate mlmeautolevel */
	if ((priv->mib.iw_mode == IW_MODE_MASTER) &&
	    (acl->policy != MAC_POLICY_OPEN))
		priv->mib.mlmeautolevel = DOT11_MLME_INTERMEDIATE;
	else
		priv->mib.mlmeautolevel = CARD_DEFAULT_MLME_MODE;
	if (priv->wpa)
		priv->mib.mlmeautolevel = DOT11_MLME_EXTENDED;
	/* restart the card with our new policy */
	prism54_mib_init(priv);

	up(&priv->mib.sem);

	return 0;
}

static int
prism54_get_policy(struct net_device *ndev, struct iw_request_info *info,
		   __u32 * uwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	struct islpci_acl *acl = &priv->acl;

	*uwrq = acl->policy;

	return 0;
}

/* Return 1 only if client should be accepted. */

static int
prism54_mac_accept(struct islpci_acl *acl, char *mac)
{
	struct list_head *ptr;
	struct mac_entry *entry;
	int res = 0;

	if (down_interruptible(&acl->sem))
		return -ERESTARTSYS;

	if (acl->policy == MAC_POLICY_OPEN) {
		up(&acl->sem);
		return 1;
	}

	for (ptr = acl->mac_list.next; ptr != &acl->mac_list; ptr = ptr->next) {
		entry = list_entry(ptr, struct mac_entry, _list);
		if (memcmp(entry->addr, mac, ETH_ALEN) == 0) {
			res = 1;
			break;
		}
	}
	res = (acl->policy == MAC_POLICY_ACCEPT) ? !res : res;
	up(&acl->sem);

	return res;
}

static int
prism54_kick_all(struct net_device *ndev, struct iw_request_info *info,
		 struct iw_point *dwrq, char *extra)
{
	struct obj_mlme *mlme;

	mlme = kmalloc(sizeof (struct obj_mlme), GFP_KERNEL);
	if (mlme == NULL)
		return -ENOMEM;

	/* Tell the card to kick every client */
	mlme->id = 0;
	MGT_SET_REQUEST_NOCACHE_NORET(ndev->priv, DOT11_OID_DISASSOCIATE,
				      mlme, sizeof (struct obj_mlme));
	kfree(mlme);

	return 0;
}

static int
prism54_kick_mac(struct net_device *ndev, struct iw_request_info *info,
		 struct sockaddr *awrq, char *extra)
{
	struct obj_mlme *mlme;
	struct sockaddr *addr = (struct sockaddr *) extra;

	if (addr->sa_family != ARPHRD_ETHER)
		return -EOPNOTSUPP;

	mlme = kmalloc(sizeof (struct obj_mlme), GFP_KERNEL);
	if (mlme == NULL)
		return -ENOMEM;

	/* Tell the card to only kick the corresponding bastard */
	memcpy(mlme->address, addr->sa_data, ETH_ALEN);
	mlme->id = -1;
	MGT_SET_REQUEST_NOCACHE_NORET(ndev->priv, DOT11_OID_DISASSOCIATE,
				      mlme, sizeof (struct obj_mlme));
	kfree(mlme);

	return 0;
}

/* Translate a TRAP oid into a wireless event. Called in islpci_mgt_receive. */

static inline void
format_event(islpci_private *priv, char *dest, const char *str,
	     const struct obj_mlme *mlme, u16 *length, int error)
{
	const u8 *a = mlme->address;
	int n = snprintf(dest, IW_CUSTOM_MAX,
			 "%s %s %2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X %s",
			 str,
			 ((priv->mib.mode == INL_MODE_CLIENT) ? "to" : "from"),
			 a[0], a[1], a[2], a[3], a[4], a[5],
			 (error ? (mlme->code ? " : REJECTED " : " : ACCEPTED ")
			  : ""));
	BUG_ON(n > IW_CUSTOM_MAX);
	*length = n;
}

static void
send_formatted_event(islpci_private *priv, const char *str,
		     const struct obj_mlme *mlme, int error)
{
	union iwreq_data wrqu;

	wrqu.data.pointer = kmalloc(IW_CUSTOM_MAX, GFP_KERNEL);
	if (!wrqu.data.pointer)
		return;
	wrqu.data.length = 0;
	format_event(priv, wrqu.data.pointer, str, mlme, &wrqu.data.length,
		     error);
	wireless_send_event(priv->ndev, IWEVCUSTOM, &wrqu, wrqu.data.pointer);
	kfree(wrqu.data.pointer);
}

static void
send_simple_event(islpci_private *priv, const char *str)
{
	union iwreq_data wrqu;
	int n = strlen(str);

	wrqu.data.pointer = kmalloc(IW_CUSTOM_MAX, GFP_KERNEL);
	if (!wrqu.data.pointer)
		return;
	BUG_ON(n > IW_CUSTOM_MAX);
	wrqu.data.length = n;
	strcpy(wrqu.data.pointer, str);
	wireless_send_event(priv->ndev, IWEVCUSTOM, &wrqu, wrqu.data.pointer);
	kfree(wrqu.data.pointer);
}

static void
link_changed(struct net_device *ndev, u32 bitrate)
{
	islpci_private *priv = ndev->priv;

	if (le32_to_cpu(bitrate)) {
		if (priv->mib.iw_mode == IW_MODE_INFRA) {
			union iwreq_data uwrq;
			prism54_get_wap(ndev, NULL, (struct sockaddr *) &uwrq,
					NULL);
			wireless_send_event(ndev, SIOCGIWAP, &uwrq, NULL);
		} else
			send_simple_event(ndev->priv, "Link established");
	} else
		send_simple_event(ndev->priv, "Link lost");
}


/* Beacon/ProbeResp payload header */
struct ieee80211_beacon_phdr {
	u8 timestamp[8];
	u16 beacon_int;
	u16 capab_info;
} __attribute__ ((packed));

#define WLAN_EID_GENERIC 0xdd
static u8 wpa_oid[4] = { 0x00, 0x50, 0xf2, 1 };

#define MAC2STR(a) (a)[0], (a)[1], (a)[2], (a)[3], (a)[4], (a)[5]
#define MACSTR "%02x:%02x:%02x:%02x:%02x:%02x"



void prism54_wpa_ie_add(islpci_private *priv, u8 *bssid,
			u8 *wpa_ie, size_t wpa_ie_len)
{
	struct list_head *ptr;
	struct islpci_bss_wpa_ie *bss = NULL;

	if (wpa_ie_len > MAX_WPA_IE_LEN)
		wpa_ie_len = MAX_WPA_IE_LEN;

	if (down_interruptible(&priv->wpa_sem))
		return;

	/* try to use existing entry */
	list_for_each(ptr, &priv->bss_wpa_list) {
		bss = list_entry(ptr, struct islpci_bss_wpa_ie, list);
		if (memcmp(bss->bssid, bssid, ETH_ALEN) == 0) {
			list_move(&bss->list, &priv->bss_wpa_list);
			break;
		}
		bss = NULL;
	}

	if (bss == NULL) {
		/* add a new BSS entry; if max number of entries is already
		 * reached, replace the least recently updated */
		if (priv->num_bss_wpa >= MAX_BSS_WPA_IE_COUNT) {
			bss = list_entry(priv->bss_wpa_list.prev,
					 struct islpci_bss_wpa_ie, list);
			list_del(&bss->list);
		} else {
			bss = kmalloc(sizeof(*bss), GFP_ATOMIC);
			if (bss != NULL) {
				priv->num_bss_wpa++;
				memset(bss, 0, sizeof(*bss));
			}
		}
		if (bss != NULL) {
			memcpy(bss->bssid, bssid, ETH_ALEN);
			list_add(&bss->list, &priv->bss_wpa_list);
		}
	}

	if (bss != NULL) {
		memcpy(bss->wpa_ie, wpa_ie, wpa_ie_len);
		bss->wpa_ie_len = wpa_ie_len;
		bss->last_update = jiffies;
	} else {
		printk(KERN_DEBUG "Failed to add BSS WPA entry for " MACSTR
		       "\n", MAC2STR(bssid));
	}

	/* expire old entries from WPA list */
	while (priv->num_bss_wpa > 0) {
		bss = list_entry(priv->bss_wpa_list.prev,
				 struct islpci_bss_wpa_ie, list);
		if (!time_after(jiffies, bss->last_update + 60 * HZ))
			break;

		list_del(&bss->list);
		priv->num_bss_wpa--;
		kfree(bss);
	}

	up(&priv->wpa_sem);
}


size_t prism54_wpa_ie_get(islpci_private *priv, u8 *bssid, u8 *wpa_ie)
{
	struct list_head *ptr;
	struct islpci_bss_wpa_ie *bss = NULL;
	size_t len = 0;

	if (down_interruptible(&priv->wpa_sem))
		return 0;

	list_for_each(ptr, &priv->bss_wpa_list) {
		bss = list_entry(ptr, struct islpci_bss_wpa_ie, list);
		if (memcmp(bss->bssid, bssid, ETH_ALEN) == 0)
			break;
		bss = NULL;
	}
	if (bss) {
		len = bss->wpa_ie_len;
		memcpy(wpa_ie, bss->wpa_ie, len);
	}
	up(&priv->wpa_sem);

	return len;
}


void prism54_wpa_ie_init(islpci_private *priv)
{
	INIT_LIST_HEAD(&priv->bss_wpa_list);
	sema_init(&priv->wpa_sem, 1);
}


void prism54_wpa_ie_clean(islpci_private *priv)
{
	struct list_head *ptr, *n;

	list_for_each_safe(ptr, n, &priv->bss_wpa_list) {
		struct islpci_bss_wpa_ie *bss;
		bss = list_entry(ptr, struct islpci_bss_wpa_ie, list);
		kfree(bss);
	}
}


static void prism54_process_bss_data(islpci_private *priv, u32 oid, u8 *addr,
				     u8 *payload, size_t len)
{
	struct ieee80211_beacon_phdr *hdr;
	u8 *pos, *end;

	if (!priv->wpa)
		return;

	hdr = (struct ieee80211_beacon_phdr *) payload;
	pos = (u8 *) (hdr + 1);
	end = payload + len;
	while (pos < end) {
		if (pos + 2 + pos[1] > end) {
			printk(KERN_DEBUG "Parsing Beacon/ProbeResp failed "
			       "for " MACSTR "\n", MAC2STR(addr));
			return;
		}
		if (pos[0] == WLAN_EID_GENERIC && pos[1] >= 4 &&
		    memcmp(pos + 2, wpa_oid, 4) == 0) {
			prism54_wpa_ie_add(priv, addr, pos, pos[1] + 2);
			return;
		}
		pos += 2 + pos[1];
	}
}


int
prism54_process_trap_helper(islpci_private *priv, u32 oid, char *data)
{
	struct obj_mlme *mlme = (struct obj_mlme *) data;
	size_t len;
	u8 *payload, *pos = (u8 *) (mlme + 1);

	len = pos[0] | (pos[1] << 8); /* little endian data length */
	payload = pos + 2;

	/* I think all trapable objects are listed here.
	 * Some oids have a EX version. The difference is that they are emitted
	 * in DOT11_MLME_EXTENDED mode (set with DOT11_OID_MLMEAUTOLEVEL)
	 * with more info.
	 * The few events already defined by the wireless tools are not really
	 * suited. We use the more flexible custom event facility.
	 */

	switch (oid) {

	case GEN_OID_LINKSTATE:
		link_changed(priv->ndev, (u32) *data);
		break;

	case DOT11_OID_MICFAILURE:
		send_simple_event(priv, "Mic failure");
		break;

	case DOT11_OID_DEAUTHENTICATE:
		send_formatted_event(priv, "DeAuthenticate request", mlme, 0);
		break;

#define MIB_STATE_OK(mib)\
   (((mib).mlmeautolevel == DOT11_MLME_INTERMEDIATE) &&    \
	((mib).mode != INL_MODE_CLIENT))

	case DOT11_OID_AUTHENTICATE:
		if ((mlme->state == DOT11_STATE_AUTHING)
		    && MIB_STATE_OK(priv->mib)) {
			/* Someone is requesting auth and we must respond. Just send back
			 * the trap with error code set accordingly.
			 */
			mlme->code =
			    prism54_mac_accept(&priv->acl,
					       mlme->address) ? 0 : 1;
			MGT_SET_REQUEST_NOCACHE(priv, oid, data,
						sizeof (struct obj_mlme));
		}
		send_formatted_event(priv, "Authenticate request", mlme, 1);
		break;

	case DOT11_OID_DISASSOCIATE:
		send_formatted_event(priv, "Disassociate request", mlme, 0);
		break;

	case DOT11_OID_ASSOCIATE:
		if ((mlme->state == DOT11_STATE_ASSOCING)
		    && MIB_STATE_OK(priv->mib)) {
			/* Someone wants to associate. We must respond */
			mlme->code =
			    prism54_mac_accept(&priv->acl,
					       mlme->address) ? 0 : 1;
			MGT_SET_REQUEST_NOCACHE(priv, oid, data,
						sizeof (struct obj_mlme));
		}
		send_formatted_event(priv, "Associate request", mlme, 1);
		break;

	case DOT11_OID_REASSOCIATE:
		if ((mlme->state == DOT11_STATE_AUTHING)
		    && MIB_STATE_OK(priv->mib)) {
			/* Someone is requesting to reauth. Respond. */
			mlme->code =
			    prism54_mac_accept(&priv->acl,
					       mlme->address) ? 0 : 1;
			MGT_SET_REQUEST_NOCACHE(priv, oid, data,
						sizeof (struct obj_mlme));
		}
		send_formatted_event(priv, "ReAssociate request", mlme, 1);
		break;

	case DOT11_OID_BEACON:
		prism54_process_bss_data(priv, oid, mlme->address,
					 payload, len);
		send_formatted_event(priv,
				     "Received a beacon from an unkown AP",
				     mlme, 0);
		break;

	case DOT11_OID_PROBE:
		/* we received a probe from a client. */
		prism54_process_bss_data(priv, oid, mlme->address,
					 payload, len);
		send_formatted_event(priv, "Received a probe from client", mlme,
				     0);
		break;

		/* Note : the following should never happen since we don't run the card in
		 * extended mode.
		 * Note : "mlme" is actually a "struct obj_mlmeex *" here, but this
		 * is backward compatible layout-wise with "struct obj_mlme".
		 */

	case DOT11_OID_DEAUTHENTICATEEX:
		send_formatted_event(priv, "DeAuthenticate request", mlme, 0);
		break;

	case DOT11_OID_AUTHENTICATEEX:
		send_formatted_event(priv, "Authenticate request", mlme, 1);
		break;

	case DOT11_OID_DISASSOCIATEEX:
		send_formatted_event(priv, "Disassociate request", mlme, 0);
		break;

	case DOT11_OID_ASSOCIATEEX:
		send_formatted_event(priv, "Associate request", mlme, 1);
		break;

	case DOT11_OID_REASSOCIATEEX:
		send_formatted_event(priv, "Reassociate request", mlme, 1);
		break;

	default:
		return -EINVAL;
	}

	return 0;
}

/*
 * Process a device trap.  This is called via schedule_work(), outside of
 * interrupt context, no locks held.
 */
void
prism54_process_trap(void *data)
{
	struct islpci_mgmtframe *frame = data;

	prism54_process_trap_helper(frame->ndev->priv, frame->header->oid,
                                    frame->data);
	islpci_mgt_release(frame);
}

int
prism54_set_mac_address(struct net_device *ndev, void *addr)
{
	islpci_private *priv = ndev->priv;

	if (ndev->addr_len != 6)
		return -1;

	MGT_SET_REQUEST(priv, GEN_OID_MACADDRESS,
			&((struct sockaddr *) addr)->sa_data, 6, macaddress);

	return 0;
}

int
prism54_ioctl(struct net_device *ndev, struct ifreq *rq, int cmd)
{
	/* should we really support this old stuff ? */
	return -EOPNOTSUPP;
}


int
prism54_set_wpa(struct net_device *ndev, struct iw_request_info *info,
		__u32 *uwrq, char *extra)
{
	islpci_private *priv = ndev->priv;

	if (down_interruptible(&priv->mib.sem))
		return -ERESTARTSYS;

	priv->wpa = *uwrq;
	if (priv->wpa)
		priv->mib.mlmeautolevel = DOT11_MLME_EXTENDED;
	prism54_mib_init(priv);

	up(&priv->mib.sem);

	return 0;
}


int
prism54_get_wpa(struct net_device *ndev, struct iw_request_info *info,
		__u32 *uwrq, char *extra)
{
	islpci_private *priv = ndev->priv;
	*uwrq = priv->wpa;
	return 0;
}

#if WIRELESS_EXT > 12

static const iw_handler prism54_handler[] = {
	(iw_handler) prism54_commit,	/* SIOCSIWCOMMIT */
	(iw_handler) prism54_get_name,	/* SIOCGIWNAME */
	(iw_handler) NULL,	/* SIOCSIWNWID */
	(iw_handler) NULL,	/* SIOCGIWNWID */
	(iw_handler) prism54_set_freq,	/* SIOCSIWFREQ */
	(iw_handler) prism54_get_freq,	/* SIOCGIWFREQ */
	(iw_handler) prism54_set_mode,	/* SIOCSIWMODE */
	(iw_handler) prism54_get_mode,	/* SIOCGIWMODE */
	(iw_handler) prism54_set_sens,	/* SIOCSIWSENS */
	(iw_handler) prism54_get_sens,	/* SIOCGIWSENS */
	(iw_handler) NULL,	/* SIOCSIWRANGE */
	(iw_handler) prism54_get_range,	/* SIOCGIWRANGE */
	(iw_handler) NULL,	/* SIOCSIWPRIV */
	(iw_handler) NULL,	/* SIOCGIWPRIV */
	(iw_handler) NULL,	/* SIOCSIWSTATS */
	(iw_handler) NULL,	/* SIOCGIWSTATS */
#if WIRELESS_EXT > 15
	iw_handler_set_spy,	/* SIOCSIWSPY */
	iw_handler_get_spy,	/* SIOCGIWSPY */
	iw_handler_set_thrspy,	/* SIOCSIWTHRSPY */
	iw_handler_get_thrspy,	/* SIOCGIWTHRSPY */
#else				/* WIRELESS_EXT > 15 */
	(iw_handler) NULL,	/* SIOCSIWSPY */
	(iw_handler) NULL,	/* SIOCGIWSPY */
	(iw_handler) NULL,	/* -- hole -- */
	(iw_handler) NULL,	/* -- hole -- */
#endif				/* WIRELESS_EXT > 15 */
	(iw_handler) prism54_set_wap,	/* SIOCSIWAP */
	(iw_handler) prism54_get_wap,	/* SIOCGIWAP */
	(iw_handler) NULL,	/* -- hole -- */
	(iw_handler) NULL,	/* SIOCGIWAPLIST depreciated */
#if WIRELESS_EXT > 13
	(iw_handler) prism54_set_scan,	/* SIOCSIWSCAN */
	(iw_handler) prism54_get_scan,	/* SIOCGIWSCAN */
#else				/* WIRELESS_EXT > 13 */
	(iw_handler) NULL,	/* SIOCSIWSCAN */
	(iw_handler) NULL,	/* SIOCGIWSCAN */
#endif				/* WIRELESS_EXT > 13 */
	(iw_handler) prism54_set_essid,	/* SIOCSIWESSID */
	(iw_handler) prism54_get_essid,	/* SIOCGIWESSID */
	(iw_handler) prism54_set_nick, /* SIOCSIWNICKN */
	(iw_handler) prism54_get_nick,	/* SIOCGIWNICKN */
	(iw_handler) NULL,	/* -- hole -- */
	(iw_handler) NULL,	/* -- hole -- */
	(iw_handler) NULL,	/* SIOCSIWRATE */
	(iw_handler) prism54_get_rate,	/* SIOCGIWRATE */
	(iw_handler) prism54_set_rts,	/* SIOCSIWRTS */
	(iw_handler) prism54_get_rts,	/* SIOCGIWRTS */
	(iw_handler) prism54_set_frag,	/* SIOCSIWFRAG */
	(iw_handler) prism54_get_frag,	/* SIOCGIWFRAG */
	(iw_handler) prism54_set_txpower,	/* SIOCSIWTXPOW */
	(iw_handler) prism54_get_txpower,	/* SIOCGIWTXPOW */
	(iw_handler) prism54_set_retry,	/* SIOCSIWRETRY */
	(iw_handler) prism54_get_retry,	/* SIOCGIWRETRY */
	(iw_handler) prism54_set_encode,	/* SIOCSIWENCODE */
	(iw_handler) prism54_get_encode,	/* SIOCGIWENCODE */
	(iw_handler) NULL,	/* SIOCSIWPOWER */
	(iw_handler) NULL,	/* SIOCGIWPOWER */
};

/* The low order bit identify a SET (0) or a GET (1) ioctl.  */

#define PRISM54_RESET		SIOCIWFIRSTPRIV
#define PRISM54_GET_BEACON	SIOCIWFIRSTPRIV+1
#define PRISM54_SET_BEACON	SIOCIWFIRSTPRIV+2
#define PRISM54_GET_POLICY SIOCIWFIRSTPRIV+3
#define PRISM54_SET_POLICY SIOCIWFIRSTPRIV+4
#define PRISM54_GET_MAC 	   SIOCIWFIRSTPRIV+5
#define PRISM54_ADD_MAC 	   SIOCIWFIRSTPRIV+6

#define PRISM54_DEL_MAC    SIOCIWFIRSTPRIV+8

#define PRISM54_KICK_MAC   SIOCIWFIRSTPRIV+10

#define PRISM54_KICK_ALL   SIOCIWFIRSTPRIV+12

#define PRISM54_GET_WPA	   SIOCIWFIRSTPRIV+13
#define PRISM54_SET_WPA	   SIOCIWFIRSTPRIV+14

static const struct iw_priv_args prism54_private_args[] = {
/*{ cmd, set_args, get_args, name } */
	{PRISM54_RESET, 0, 0, "reset"},
	{PRISM54_GET_BEACON, 0, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	 "getBeaconPeriod"},
	{PRISM54_SET_BEACON, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0,
	 "setBeaconPeriod"},
	{PRISM54_GET_POLICY, 0, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	 "getPolicy"},
	{PRISM54_SET_POLICY, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0,
	 "setPolicy"},
	{PRISM54_GET_MAC, 0, IW_PRIV_TYPE_ADDR | 64, "getMac"},
	{PRISM54_ADD_MAC, IW_PRIV_TYPE_ADDR | IW_PRIV_SIZE_FIXED | 1, 0,
	 "addMac"},
	{PRISM54_DEL_MAC, IW_PRIV_TYPE_ADDR | IW_PRIV_SIZE_FIXED | 1, 0,
	 "delMac"},
	{PRISM54_KICK_MAC, IW_PRIV_TYPE_ADDR | IW_PRIV_SIZE_FIXED | 1, 0,
	 "kickMac"},
	{PRISM54_KICK_ALL, 0, 0, "kickAll"},
	{ PRISM54_GET_WPA, 0, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
	  "get_wpa" },
	{ PRISM54_SET_WPA, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0,
	  "set_wpa" },
};

static const iw_handler prism54_private_handler[] = {
	(iw_handler) prism54_reset,
	(iw_handler) prism54_get_beacon,
	(iw_handler) prism54_set_beacon,
	(iw_handler) prism54_get_policy,
	(iw_handler) prism54_set_policy,
	(iw_handler) prism54_get_mac,
	(iw_handler) prism54_add_mac,
	(iw_handler) NULL,
	(iw_handler) prism54_del_mac,
	(iw_handler) NULL,
	(iw_handler) prism54_kick_mac,
	(iw_handler) NULL,
	(iw_handler) prism54_kick_all,
	(iw_handler) prism54_get_wpa,
	(iw_handler) prism54_set_wpa,
};

const struct iw_handler_def prism54_handler_def = {
	.num_standard = sizeof (prism54_handler) / sizeof (iw_handler),
	.num_private = sizeof (prism54_private_handler) / sizeof (iw_handler),
	.num_private_args =
	    sizeof (prism54_private_args) / sizeof (struct iw_priv_args),
	.standard = (iw_handler *) prism54_handler,
	.private = (iw_handler *) prism54_private_handler,
	.private_args = (struct iw_priv_args *) prism54_private_args,
};

#endif				/* WIRELESS_EXT > 12 */
