/*  $Header: /var/lib/cvs/prism54-ng/ksrc/islpci_hotplug.c,v 1.19 2003/11/12 08:08:55 hvr Exp $
 *  
 *  Copyright (C) 2002 Intersil Americas Inc.
 *  Copyright (C) 2003 Herbert Valerio Riedel <hvr@gnu.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/delay.h>

#include "version_info.h"
#include "isl_gen.h"
#include "isl_38xx.h"
#include "islpci_dev.h"
#include "islpci_mgt.h"

#ifdef INTERSIL_EVENTS
#include <bloboid.h>
#include "isl_mgt.h"
#endif

/******************************************************************************
*        Global variable definition section
******************************************************************************/
MODULE_AUTHOR
    ("W.Termorshuizen, R.Bastings, Herbert Valerio Riedel <hvr@hvrlab.org>, prism54.org team");
MODULE_DESCRIPTION("Intersil 802.11 Wireless LAN adapter");
MODULE_LICENSE("GPL");

#define DRIVER_NAME		"prism54"

static const struct pci_device_id pci_id_table[] = {
	{
	 PCIVENDOR_INTERSIL, PCIDEVICE_ISL3877,
	 PCI_ANY_ID, PCI_ANY_ID, 0, 0,
	 /* Driver data, we just put the name here */
	 (unsigned long) "Intersil PRISM Indigo Wireless LAN adapter"},
	{
	 PCIVENDOR_INTERSIL, PCIDEVICE_ISL3890,
	 PCI_ANY_ID, PCI_ANY_ID, 0, 0,
	 /* Driver data, we just put the name here */
	 (unsigned long) "Intersil PRISM Duette Wireless LAN adapter"},
	{
	 PCIVENDOR_3COM, PCIDEVICE_3COM6001,
	 PCI_ANY_ID, PCI_ANY_ID, 0, 0,
	 (unsigned long) "3COM 3CRWE154G72 Wireless LAN adapter"},
	{
	 0, 0, 0, 0, 0, 0, 0}
};

/* register the device with the Hotplug facilities of the kernel */
MODULE_DEVICE_TABLE(pci, pci_id_table);

static int islpci_probe_pci(struct pci_dev *, const struct pci_device_id *);
static void islpci_remove_pci(struct pci_dev *);
static int islpci_suspend(struct pci_dev *, u32 state);
static int islpci_resume(struct pci_dev *);

struct pci_driver islpci_pci_drv_id = {
	.name = DRIVER_NAME,	/* Driver name */
	.id_table = pci_id_table,	/* id table */
	.probe = islpci_probe_pci,	/* probe function */
	.remove = islpci_remove_pci,	/* remove function */
	.suspend = islpci_suspend,
	.resume = islpci_resume,
	/* .enable_wake ; we don't support this yet */
};

/******************************************************************************
    Module initialization functions
******************************************************************************/

int
islpci_probe_pci(struct pci_dev *pdev, const struct pci_device_id *id)
{
	struct net_device *ndev;
	u8 latency_tmr;
	islpci_private *priv;
	int rvalue;
	int dma_mask = 0xffffffff;

	/* TRACE(DRIVER_NAME); */

	/* Enable the pci device */
	if (pci_enable_device(pdev)) {
		printk(KERN_ERR "%s: pci_enable_device() failed.\n",
		       DRIVER_NAME);
		return -EIO;
	}

	/* check whether the latency timer is set correctly */
	pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &latency_tmr);
#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "latency timer: %x\n", latency_tmr);
#endif
	if (latency_tmr < PCIDEVICE_LATENCY_TIMER_MIN) {
		/* set the latency timer */
		pci_write_config_byte(pdev, PCI_LATENCY_TIMER,
				      PCIDEVICE_LATENCY_TIMER_VAL);

	}
	/* determine what the supported DMA memory region is */
	while (pci_set_dma_mask(pdev, dma_mask) != 0) {
		/* range not supported, shift the mask and check again */
		if (dma_mask >>= 1, dma_mask == 0) {
			/* mask is zero, DMA memory not supported by PCI */
			printk(KERN_ERR "%s: DMA Memory not supported\n",
			       DRIVER_NAME);
			goto do_pci_disable_device;
		}
	}

#if VERBOSE > SHOW_ERROR_MESSAGES
	DEBUG(SHOW_TRACING, "DMA Memory support mask is 0x%x \n", dma_mask);
#endif

	/* request the pci device I/O regions */
	if (rvalue = pci_request_regions(pdev, DRIVER_NAME), rvalue) {
		printk(KERN_ERR "%s: pci_request_regions failure (rc=%d)\n",
		       DRIVER_NAME, rvalue);
		goto do_pci_disable_device;
	}
	/* enable PCI bus-mastering */
	DEBUG(SHOW_TRACING, "%s: pci_set_master(pdev)\n", DRIVER_NAME);
	pci_set_master(pdev);

	/* setup the network device interface and its structure */
	if (!(ndev = islpci_setup(pdev))) {
		/* error configuring the driver as a network device */
		printk(KERN_ERR "%s: could not configure network device\n",
		       DRIVER_NAME);
		goto do_pci_release_regions;
	}
#ifdef WDS_LINKS
	mgt_indication_handler(DEV_NETWORK, ndev->ifindex, DOT11_OID_WDSLINKADD,
			       islpci_wdslink_add_hndl);
	mgt_indication_handler(DEV_NETWORK, ndev->ifindex,
			       DOT11_OID_WDSLINKREMOVE,
			       islpci_wdslink_del_hndl);
#endif

	priv = ndev->priv;

	/* request for the interrupt before uploading the firmware */

	rvalue = request_irq(pdev->irq, &islpci_interrupt,
			     SA_INTERRUPT | SA_SHIRQ, ndev->name, priv);

	if (rvalue) {
		/* error, could not hook the handler to the irq */
		printk(KERN_ERR "%s: could not install IRQ handler\n",
		       ndev->name);
		goto do_unregister_netdev;
	}

	/* reset data structures, upload firmware and reset device */
	if (islpci_reset(priv))
		goto do_free_irq;

	return 0;

      do_free_irq:
	free_irq(pdev->irq, priv);
      do_unregister_netdev:
	unregister_netdev(ndev);
	islpci_free_memory(priv);
	pci_set_drvdata(pdev, 0);
	free_netdev(ndev);
	priv = 0;
      do_pci_release_regions:
	pci_release_regions(pdev);
      do_pci_disable_device:
	pci_disable_device(pdev);
	return -EIO;
}

/* set by cleanup_module */
static volatile int __in_cleanup_module = 0;

/* this one removes one(!!) instance only */
void
islpci_remove_pci(struct pci_dev *pdev)
{
	struct net_device *ndev = pci_get_drvdata(pdev);
	islpci_private *priv = ndev ? ndev->priv : 0;
	BUG_ON(!priv);

	if (!__in_cleanup_module) {
		printk(KERN_DEBUG "%s: hot unplug detected", ndev->name);
		priv->hw_unavailable++;
	}

	printk(KERN_DEBUG "%s: removing device\n", ndev->name);

	unregister_netdev(ndev);

	/* free the interrupt request */

	if (!priv->hw_unavailable) {
		isl38xx_disable_interrupts(priv->device_base);
		priv->hw_unavailable++;
		pci_set_power_state(pdev, 3);	/* try to power-off */
	}

	free_irq(pdev->irq, priv);

	/* free the PCI memory and unmap the remapped page */
	islpci_free_memory(priv);

	pci_set_drvdata(pdev, 0);
	free_netdev(ndev);
	priv = 0;

	pci_release_regions(pdev);

	pci_disable_device(pdev);
}

int
islpci_suspend(struct pci_dev *pdev, u32 state)
{
	struct net_device *ndev = pci_get_drvdata(pdev);
	islpci_private *priv = ndev ? ndev->priv : 0;
	BUG_ON(!priv);

	printk(KERN_NOTICE "%s: got suspend request (state %d)\n",
	       ndev->name, state);

	pci_save_state(pdev, priv->pci_state);

	/* tell the device not to trigger interrupts for now... */
	isl38xx_disable_interrupts(priv->device_base);

	/* from now on assume the hardware was already powered down
	   and don't touch it anymore */
	priv->hw_unavailable++;

	netif_stop_queue(ndev);
	netif_device_detach(ndev);

	return 0;
}

int
islpci_resume(struct pci_dev *pdev)
{
	struct net_device *ndev = pci_get_drvdata(pdev);
	islpci_private *priv = ndev ? ndev->priv : 0;
	BUG_ON(!priv);

	printk(KERN_NOTICE "%s: got resume request\n", ndev->name);

	pci_restore_state(pdev, priv->pci_state);

	/* from now on we have access to the registers again */
	priv->hw_unavailable++;

	islpci_reset(priv);

	netif_device_attach(ndev);
	netif_start_queue(ndev);

	return 0;
}

int
init_module(void)
{
	printk(KERN_INFO "Loaded %s driver, version %s\n",
	       DRIVER_NAME, VERSIONID);

#ifdef INTERSIL_EVENTS
	mgt_init();
#endif

	if (pci_register_driver(&islpci_pci_drv_id) <= 0) {
		printk(KERN_ERR "%s: No devices found, driver not "
		       "installed.\n", DRIVER_NAME);
		pci_unregister_driver(&islpci_pci_drv_id);
		return -ENODEV;
	}
#ifdef INTERSIL_EVENTS
	mgt_indication_handler(DEV_NETWORK, 0, 0, islpci_mgt_indication);
#endif

	return 0;
}

/* when this is called, all instances have already been destroyed by
 * islpci_remove_pci calls */

void
cleanup_module(void)
{
#ifdef INTERSIL_EVENTS
	mgt_cleanup();
#endif
	__in_cleanup_module = 1;

	pci_unregister_driver(&islpci_pci_drv_id);

	printk(KERN_INFO "Unloaded %s driver\n", DRIVER_NAME);
}
