#!/bin/bash
# Author: Luis R. Rodriguez <mcgrof@ruslug.rutgers.edu>
# Script to automatically generate patches for 2.4 and/or 2.6
# against latest cvs repository
#
# You need at least 2.0 GB hard drive space to make this work ;)
#
# Requirements:
# You must have: a /home/$USER/kernel-src/ dir 
# with all the latest kernels, correctly
# renamed, so if you have linux-2.4.22 patched 
# up to 2.4.23-pre8, you should rename it to 
# that. So at the time of the writing of this script, 
# for example, you should have:
# linux-2.4.22
# linux-2.4.23-pre8
# linux-2.6-0-test9
#
# Also, you will need a VANILLA_BACKUPS_DIR to 
# keep *really* fresh, untouched 
# sources if you want to use the --purge option. 
# In that dir you need all the kernel
# sources mentioned above
# 
# You also should have the dontdiff file in your /home/$USER/lib/ dir
# If you don't have it you can steal it off of me:
# http://prism54.org/~mcgrof/dontdiff

# Who are you?
USER="mcgrof"
# Kernel prefix
KPRE="linux"
# Driver/project name
DRV="prism54"
# Location of kernel sources
KSRC_DIR="/home/$USER/kernel-src/"
# Location of vanilla source backups
VANILLA_BACKUPS_DIR="$KSRC_DIR/vanilla-backups/"
# Location of where CVS modules go
CVS_DIR="/home/$USER/tmp/patchomatic"
# Name of this project's CVS module
CVS_MODULE="prism54-ng"
# Prism54 driver code destination on kernel tree
PRISM54_KERN_DEST="drivers/net/wireless/prism54"
# Be smart, diff only the following dirs/files:
# Generic:
# 2.4 specific files
DIFF_LIST_24="Documentation/Configure.help drivers/net/wireless/"
# 2.6 specific files
DIFF_LIST_26="drivers/net/wireless/"
# Kernel.org finger banner
FINGER_URL="kernel.org/kdist/finger_banner"
# File to store finger from kernel
TMP="/tmp/finger.tmp"
# contact to send new kernel release notifications to
#CONTACT="prism54-users@prism54.org" # Once patchomatic.sh is stable
#I'm messing the line bellow just because tons of people actually have sent
#me e-mail -- that is, they don't edit this!
#CONTACT="mcgrof@ruslug.rutgers.edu"
CONTACT="root@localhost"
# File used by notify_new_release to see if we've already have
# sent out an e-mail notification
MAIL_SENT="${KSRC_DIR}MAIL_SENT"
LINE="----------------------------------------------------------------"

# Colors :D
# Text Colors
RED="\e[31m"
GREEN="\e[32m"
YELLOW="\e[33m"
BLUE="\e[34m"
CYAN="\e[36m"
WHITE="\e[37m"
NO_COLOR="\e[0m"

# Standard strings
K_24_STRING="The latest 2.4 version of the Linux kernel is"
STABLE_STRING="The latest stable version of the Linux kernel is"
STABLE_PRE_STRING="The latest prepatch for the stable Linux kernel tree is"
BETA_STRING="The latest beta version of the Linux kernel is"

lynx --dump $FINGER_URL > $TMP

# Versioning
K_24=`grep "$K_24_STRING"	$TMP | awk -F: '{print $2}' | awk '{print $1}'`
STABLE=`grep "$STABLE_STRING"	$TMP | awk -F: '{print $2}' | awk '{print $1}'`
PRE=`grep "$STABLE_PRE_STRING"	$TMP | awk -F: '{print $2}' | awk '{print $1}'`
BETA=`grep "$BETA_STRING"	$TMP | awk -F: '{print $2}' | awk '{print $1}'`

rm -f $TMP

VANILLA_24="$KPRE-$K_24"
VANILLA_STABLE="$KPRE-$STABLE"
VANILLA_PRE="$KPRE-$PRE"
VANILLA_BETA="$KPRE-$BETA"

PRISM54_24="$VANILLA_24-$DRV"
PRISM54_STABLE="$VANILLA_STABLE-$DRV"
PRISM54_PRE="$VANILLA_PRE-$DRV"
PRISM54_BETA="$VANILLA_BETA-$DRV"

P24_LATEST="patch-2.4-$DRV-cvs-latest.bz2"
P26_LATEST="patch-2.6-$DRV-cvs-latest.bz2"

# Kernels to check for and to patch
ALL_KERNELS="$VANILLA_24 $VANILLA_STABLE"
# All directories that must exist
ALL_DIRS="$ALL_KERNELS $PRISM54_24 $PRISM54_STABLE" 

# Later on add support to check for latest version auto download/etc
versions_are_updated()
{
if [ ${#VANILLA_STABLE} -le 6 ]; then
	echo -e "Unsuccessful connecting to kernel.org"
	exit
fi
cd $KSRC_DIR
STATUS=0 # true
for i in $ALL_DIRS; do
	if [ -d $i ]; then
		echo -e "${GREEN}FOUND\t\t${i}${NO_COLOR}"
	else
		echo -e "${RED}NOT FOUND\t${i}${NO_COLOR}"
		STATUS=1
	fi
done
echo -e "$LINE"
# If archive is updated, remove any old notices
if [ $STATUS -eq 0 ]; then
	rm -f ${MAIL_SENT}
fi
return $STATUS
}

notify_new_release()
{
MESSAGE="[This is an automated message brought to you by prism54 patchomatic.sh \n
 It is generated every time kernel.org has a new kernel release]\n\n
Latest kernel releases:\n\n
Latest stable:\t\t$VANILLA_STABLE\n
Latest pre-patch:\t$VANILLA_PRE\n
Latest beta:\t\t$VANILLA_BETA\n\n\n
New Kernel patches for them will be available as soon as mcgrof updates his tree.\n
Please be patient ;)
"
SUBJECT="[New Kernel Release] Kernel.org seem to have a new release"

# Only send if we did connect to kernel.org, we can check that by
# checking the length of the variables. If we didn't connect
# all they contain is "linux-"
if [ ${#VANILLA_STABLE} -gt 6 ]; then
	if [ ! -f ${MAIL_SENT} ]; then
		echo -e $MESSAGE | mailx -s "$SUBJECT" $CONTACT
		echo -e "E-mail notification sent"
		touch ${MAIL_SENT}
	else 
		echo -e "E-mail message informing of new kernel already sent. Not resending."
	fi
else 
	echo -e "No e-mail sent due to failure in connecting to kernel.org"
fi	
}

build_prism54_dirs()
{
cd $KSRC_DIR
echo -e "Removing old data..."
for i in $ALL_KERNELS; do
	echo -e "${RED}Removing${NO_COLOR} old prism54 source from $i-$DRV..."
	
	if [ `expr $i : $KPRE'-2.4'` -gt 0 ]; then
		rm -rf $i-$DRV/drivers/net/wireless/prism54
		cp $i/drivers/net/wireless/Config.in $i-$DRV/drivers/net/wireless/Config.in
		cp $i/drivers/net/wireless/Makefile $i-$DRV/drivers/net/wireless/Makefile
		cp $i/Documentation/Configure.help $i-$DRV/Documentation/Configure.help
	# Else... if kernel is 2.6
	elif [ `expr $i : $KPRE-'2.6'` -gt 0 ]; then
		rm -rf $i-$DRV/drivers/net/wireless/prism54
		cp $i/drivers/net/wireless/Makefile $i-$DRV/drivers/net/wireless/Makefile
		cp $i/drivers/net/wireless/Kconfig $i-$DRV/drivers/net/wireless/Kconfig
	fi
done
echo -e "$LINE"
}

# Get latest source
check_out_cvs()
{
mkdir -p $CVS_DIR
echo -e "Getting latest CVS $DRV source..."
cd $CVS_DIR
if [ -d $CVS_MODULE ]; then
	cd $CVS_MODULE
	cvs update
else
	cvs co $CVS_MODULE
fi
echo -e "$LINE"
}

inject_prism54_module()
{
cd $KSRC_DIR
for i in $ALL_KERNELS; do
	mkdir -p $i-$DRV/$PRISM54_KERN_DEST
	# Inject code
	echo -e "${GREEN}Injecting${NO_COLOR} prism54 code to $i-$DRV..."
	cp $CVS_DIR/$CVS_MODULE/ksrc/*.[ch] $i-$DRV/$PRISM54_KERN_DEST
	# Inject Documenation entry

	# Deal with kernel specific stuff:
	# If kernel is 2.4 
	if [ `expr $i : $KPRE-'2.4'` -gt 0 ]; then
		echo -e "Applying prism54 2.4 patch to kernel..."
		cp $CVS_DIR/$CVS_MODULE/ksrc/Makefile.k24 $i-$DRV/$PRISM54_KERN_DEST/Makefile
		cd $i-$DRV
		patch -p1 < $CVS_DIR/$CVS_MODULE/patch-2.4-$DRV
		cd ..
	# else if kernel is 2.6
	elif [ `expr $i : $KPRE-'2.6'` -gt 0 ]; then
		echo -e "Applying prism54 2.6 patch to kernel..."
		cp $CVS_DIR/$CVS_MODULE/ksrc/Makefile.k26 $i-$DRV/$PRISM54_KERN_DEST/Makefile
		cd $i-$DRV
		patch -p1 < $CVS_DIR/$CVS_MODULE/patch-2.6-$DRV
		cd ..
	fi
done
echo -e "$LINE"
}

make_patches()
{
cd $KSRC_DIR
echo -e "${GREEN}Creating${NO_COLOR} patches now..."
for i in $ALL_KERNELS; do
	DT=`date +%Y%m%d`
	DT_DIR=`date +%Y-%m`
	mkdir -p patches/v2.4/$DT_DIR patches/v2.6/$DT_DIR
	PATCH_FILENAME="patch-${i#${KPRE}-}-$DRV-cvs$DT"
	rm -f patches/${PATCH_FILENAME}*
	echo -e "Creating $PATCH_FILENAME..."

	# Clear old patch, if exists
	rm -f $PATCH_FILENAME
	touch $PATCH_FILENAME
	# Kernel specific diffs
	if [ `expr $i : $KPRE-'2.4'` -gt 0 ]; then
	for y in $DIFF_LIST_24; do
diff -Naur -X /home/$USER/lib/dontdiff $i/$y $i-$DRV/$y >> $PATCH_FILENAME
	done
	# else if kernel is 2.6
	elif [ `expr $i : $KPRE-'2.6'` -gt 0 ]; then
	for y in $DIFF_LIST_26; do
diff -Naur -X /home/$USER/lib/dontdiff $i/$y $i-$DRV/$y >> $PATCH_FILENAME
	done
	fi
	# Deal with kernel specific moving stuff:
	# If kernel is 2.4 
	if [ `expr $i : $KPRE-'2.4'` -gt 0 ]; then
		mv $PATCH_FILENAME patches/v2.4/$DT_DIR/
		rm -f patches/v2.4/$DT_DIR/$PATCH_FILENAME.bz2
		bzip2 patches/v2.4/$DT_DIR/$PATCH_FILENAME
		cp -f patches/v2.4/$DT_DIR/$PATCH_FILENAME.bz2 patches/v2.4/$P24_LATEST
	# else if kernel is 2.6
	elif [ `expr $i : $KPRE-'2.6'` -gt 0 ]; then
		mv $PATCH_FILENAME patches/v2.6/$DT_DIR/
		rm -f patches/v2.6/$DT_DIR/$PATCH_FILENAME.bz2
		bzip2 patches/v2.6/$DT_DIR/$PATCH_FILENAME
		cp -f patches/v2.6/$DT_DIR/$PATCH_FILENAME.bz2 patches/v2.6/$P26_LATEST
	fi
done
chmod -R 755 $KSRC_DIR/patches
echo -e "$LINE"
}

# This sucker here cleans out your KSRC with fresh vanilla kernel sources
# Note: you must have the SPARE_VANILLA sources to use this ;)
purge_dirs()
{
cd $KSRC_DIR
if [ -d $VANILLA_BACKUPS_DIR ]; then
	echo -e "We're starting from scratch, this may take a while..."
	for i in $VANILLA_STABLE $VANILLA_PRE $VANILLA_BETA; do
		echo -e "Dumping old $i and $i-$DRV dirs..."
		rm -rf $i $i-$DRV
		echo -e "Starting up a fresh $i and $i-$DRV dirs..."
		cp -a $VANILLA_BACKUPS_DIR/$i $i
		cp -a $VANILLA_BAKCUPS_DIR/$i $i-$DRV
	done
	echo -e "$LINE"
fi
}

show_usage()
{
echo -e "Usage:\t$0 [--purge]"
}

# ------------------------------------------------------
# Main MAIN main __init  ;)

# If versions are not updated from latest kernel sources available
versions_are_updated 
UPDATED=$?
if [ "$UPDATED" -eq "1" ]; then
	# In the future we should get them automagically
	echo -e "${GREEN}Go get the latest source tarballs!${NO_COLOR}"
	notify_new_release 2>&1 > /dev/null
	exit -1
fi
# Check args
if [ $# -gt 1 ];  then 
	show_usage
	exit 1
elif [ $# -eq 1 ]; then
	# Deal with the arguments here
	case $1 in 
		--purge)
		purge_dirs
		;;
		*)
		show_usage
		exit 1;;
	esac
else
	check_out_cvs 2>&1 > /dev/null
	build_prism54_dirs
	inject_prism54_module
	make_patches
	echo
fi

